from loguru import logger
from sweepai.core.chat import Function
from sweepai.utils.diff import format_contents

modify_file_function = Function(
    name="modify_file",
    description="Edits the code in a file. Use start_line and end_line to completely cover the line indexes of code that should be replaced. Indent and format the code in the edits. Output the code in the order it should appear in the file. Make sure start_line and end_line do not overlap between code edits.",
    parameters={
        "type": "object",
        "properties": {
            "file_name": {
                "type": "string",
                "description": "The name of the file to modify."
            },
            "code_edits": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "start_line": {
                            "type": "integer",
                            "description": "The index where the code should start being inserted/replaced."
                        },
                        "end_line": {
                            "type": "integer",
                            "description": "The index where the code should stop being inserted/replaced. Add 1 to this number to include the line."
                        },
                        "inserted_code": {
                            "type": "string",
                            "description": "Only the new code to insert into the file. Indent and format this code properly using spaces, keeping in mind the entire block will be affected by num_indents. To delete a line, set this to '' (single quoted empty string)."
                        },
                        "num_indents": {
                            "type": "integer",
                            "description": "Use this to indent the entire inserted_code. BE SURE to match the indentation to be inline. There will be two spaces for however many num_indents are set. num_indents can be set to 0, but ONLY IF NEEDED. When it is ambiguous, set as many num_indents as possible. "
                        }
                    },
                    "required": ["start_line", "end_line", "code", "num_indents"]
                },
                "description": "An array of edits. Each `code_edit` represents a slice of the code split by newlines and delimited by `start_line` and `end_line`. Both `start_line` and `end_line` are zero-indexed and inclusive."
            }
        },
        "required": ["file_name", "code_edits"]
        }
)

def apply_code_edits(file_contents, code_edits):
    modifications = []
    for edit in code_edits:
        start_line = int(edit['start_line'])
        end_line = int(edit['end_line'])
        new_code = format_contents(edit['inserted_code'])
        # Indentation
        indentation = int(edit['num_indents'])
        logger.info(f"The code {new_code} has {indentation} indents")
        # Starts with or ends with "" should be swapped to '' for json
        if len(new_code) >= 2 and new_code[:1] == '""':
            new_code = "''" + new_code[2:]
        elif len(new_code) >= 2 and new_code[-2:] == '""':
            new_code = new_code[:-2] + "''"
        new_code = edit['inserted_code'].split('\n')
        modifications.append((start_line, end_line, new_code, indentation))

    # Sort modifications by start line in reverse order
    modifications.sort(key=lambda x: x[0], reverse=True)
    lines = file_contents.split('\n')
    for start_line, end_line, new_code, indentation in modifications:
        if start_line > end_line:
            logger.error(f"Start line {start_line} is greater than end line {end_line}")
            continue
        if start_line < 0:
            logger.error(f"Start line {start_line} is less than 0")
            continue
        if end_line > len(lines) - 1:
            logger.error(f"End line {end_line} is greater than the number of lines in the file {len(lines)}")
            continue
        # Handle duplicate lines between the existing code and new code
        indents = '  ' * indentation
        if start_line > 0 and end_line < len(lines) \
            and new_code[0] == lines[start_line-1] and new_code[-1] == lines[end_line]:
            new_code = new_code[1:-1]
            new_code = [indents + line for line in new_code]
            lines[start_line:end_line] = new_code
            continue
        elif start_line > 0 and new_code[0] == lines[start_line-1]:
            new_code = new_code[1:]
            new_code = [indents + line for line in new_code]
            lines[start_line-1:end_line + 1] = new_code # Exit and merge first line
            continue
        elif end_line < len(lines) and new_code[-1] == lines[end_line]:
            new_code = new_code[:-1]
            new_code = [indents + line for line in new_code]
            lines[start_line:end_line] = new_code # Exit and merge last line
            continue
        # Check index error
        if end_line > len(lines) - 1:
            end_line = len(lines) - 1
        new_code = [indents + line for line in new_code]
        lines[start_line:end_line + 1] = new_code # Start and end are inclusive
    return '\n'.join(lines)



from copy import deepcopy
import json
import os
from typing import Generator, Iterator, Literal, Self

import modal
import openai
import anthropic
from loguru import logger
from pydantic import BaseModel
import backoff

from sweepai.core.entities import (
    Function,
    Message,
)
from sweepai.core.prompts import (
    system_message_prompt,
    system_message_issue_comment_prompt,
)
from sweepai.utils.constants import UTILS_NAME
from sweepai.utils.prompt_constructor import HumanMessagePrompt
from sweepai.core.entities import Message, Function

# TODO: combine anthropic and openai

AnthropicModel = (
    Literal["claude-v1"]
    | Literal["claude-v1.3-100k"]
    | Literal["claude-instant-v1.1-100k"]
)
OpenAIModel = Literal["gpt-3.5-turbo"] | Literal["gpt-4"] | Literal["gpt-4-32k"] | Literal["gpt-4-0613"] | Literal["gpt-4-32k-0613"] | Literal["gpt-3.5-turbo-16k-0613"]
ChatModel = OpenAIModel | AnthropicModel
model_to_max_tokens = {
    "gpt-3.5-turbo": 4096,
    "gpt-4": 8192,
    "gpt-4-0613": 8192,
    "gpt-4-32k": 32000,
    "gpt-4-32k-0613": 32000,
    "claude-v1": 9000,
    "claude-v1.3-100k": 100000,
    "claude-instant-v1.3-100k": 100000,
    "gpt-3.5-turbo-16k-0613": 16000,
}
temperature = 0.1

def format_for_anthropic(messages: list[Message]) -> str:
    if len(messages) > 1:
        new_messages: list[Message] = [Message(role="system", content=messages[0].content + "\n" + messages[1].content)]
        messages = messages[2:] if len(messages) >= 3 else []
    else:
        new_messages: list[Message] = []
    for message in messages:
        new_messages.append(message)
    return "\n".join(
        f"{anthropic.HUMAN_PROMPT if message.role != 'assistant' else anthropic.AI_PROMPT} {message.content}"
        for message in new_messages
    ) + (anthropic.AI_PROMPT if new_messages[-1].role != "assistant" else "")


class ChatGPT(BaseModel):
    messages: list[Message] = [
        Message(
            role="system",
            content=system_message_prompt,
        )
    ]
    prev_message_states: list[list[Message]] = []
    model: ChatModel = "gpt-4-32k-0613"
    human_message: HumanMessagePrompt | None = None
    file_change_paths = []

    @classmethod
    def from_system_message_content(
        cls, human_message: HumanMessagePrompt, is_reply: bool = False, **kwargs
    ) -> Self:
        if is_reply:
            system_message_content = system_message_issue_comment_prompt
        system_message_content = (
            system_message_prompt + "\n\n" + human_message.construct_prompt()
        )
        return cls(
            messages=[
                Message(role="system", content=system_message_content, key="system")
            ],
            human_message=human_message,
            **kwargs,
        )

    @classmethod
    def from_system_message_string(cls, prompt_string, **kwargs) -> Self:
        return cls(
            messages=[Message(role="system", content=prompt_string, key="system")],
            **kwargs,
        )

    def select_message_from_message_key(
        self, message_key: str, message_role: str = None
    ):
        if message_role:
            return [
                message
                for message in self.messages
                if message.key == message_key and message.role == message_role
            ][0]
        return [message for message in self.messages if message.key == message_key][0]

    def delete_messages_from_chat(self, message_key: str):
        self.messages = [
            message for message in self.messages if message.key != message_key
        ]

    def delete_file_from_system_message(self, file_path: str):
        self.human_message.delete_file(file_path)

    def get_message_content_from_message_key(
        self, message_key: str, message_role: str = None
    ):
        return self.select_message_from_message_key(
            message_key, message_role=message_role
        ).content

    def update_message_content_from_message_key(
        self, message_key: str, new_content: str, message_role: str = None
    ):
        self.select_message_from_message_key(
            message_key, message_role=message_role
        ).content = new_content

    def chat(
        self,
        content: str,
        model: ChatModel | None = None,
        message_key: str | None = None,
        functions: list[Function] = [],
        function_name: dict | None = None,
    ):
        if self.messages[-1].function_call is None:
            self.messages.append(Message(role="user", content=content, key=message_key))
        else:
            name = self.messages[-1].function_call["name"]
            self.messages.append(Message(role="function", content=content, key=message_key, name=name))
        model = model or self.model
        is_function_call = False
        if model in [args.__args__[0] for args in OpenAIModel.__args__]:
            # might be a bug here in all of this
            response = self.call_openai(model=model, functions=functions, function_name=function_name)
            if functions:
                response, is_function_call = response
                if is_function_call:
                    self.messages.append(
                        Message(role="assistant", content=None, function_call=response, key=message_key)
                    )
                    self.prev_message_states.append(self.messages)
                    return self.messages[-1].function_call
                else:
                    self.messages.append(
                        Message(role="assistant", content=response, key=message_key)
                    )
            else:
                response = self.call_openai(model=model, functions=functions)
                self.messages.append(
                    Message(role="assistant", content=response, key=message_key)
                )
        else:
            response = self.call_anthropic(model=model)
            self.messages.append(
                Message(role="assistant", content=response, key=message_key)
            )
        self.prev_message_states.append(self.messages)
        return self.messages[-1].content
    
    def call_openai(
        self, 
        model: ChatModel | None = None,
        functions: list[Function] = [],
        function_name: dict | None = None,
    ):
        if model is None:
            model = self.model
        count_tokens = modal.Function.lookup(UTILS_NAME, "Tiktoken.count")
        messages_length = sum(
            [count_tokens.call(message.content or "") for message in self.messages]
        )
        max_tokens = model_to_max_tokens[model] - int(messages_length) - 400 # this is for the function tokens
        # TODO: Add a check to see if the message is too long
        logger.info("file_change_paths" + str(self.file_change_paths))
        if len(self.file_change_paths) > 0:
            self.file_change_paths.remove(self.file_change_paths[0])
        if max_tokens < 0:
            if len(self.file_change_paths) > 0:
                pass
            else:
                raise ValueError(f"Message is too long, max tokens is {max_tokens}")
        messages_raw = "\n".join([(message.content or "") for message in self.messages])
        logger.info(f"Input to call openai:\n{messages_raw}")

        gpt_4_buffer = 800
        if int(messages_length) + gpt_4_buffer < 6000 and model == "gpt-4-32k-0613":
            model = "gpt-4-0613"
            max_tokens = model_to_max_tokens[model] - int(messages_length) - gpt_4_buffer # this is for the function tokens
        if "gpt-4" in model:
            max_tokens = min(max_tokens, 5000)
        logger.info(f"Using the model {model}, with {max_tokens} tokens remaining")
        global retry_counter
        retry_counter = 0
        if functions:
            @backoff.on_exception(
                backoff.expo,
                Exception,
                max_tries=12,
                jitter=backoff.random_jitter,

            )
            def fetch():
                global retry_counter
                retry_counter += 1
                token_sub = retry_counter * 200
                try:
                    if function_name:
                        return (
                            openai.ChatCompletion.create(
                                model=model,
                                messages=self.messages_dicts,
                                max_tokens=max_tokens - token_sub,
                                temperature=temperature,
                                functions=[json.loads(function.json()) for function in functions],
                                function_call=function_name,
                            )
                            .choices[0].message
                        )
                    else:
                        return (
                        openai.ChatCompletion.create(
                            model=model,
                            messages=self.messages_dicts,
                            max_tokens=max_tokens - token_sub,
                            temperature=temperature,
                            functions=[json.loads(function.json()) for function in functions],
                        )
                        .choices[0].message
                    )
                except Exception as e:
                    logger.warning(e)
                    raise e
            result = fetch()
            if "function_call" in result:
                result = dict(result["function_call"]), True
            else:
                result = result["content"], False
            logger.info(f"Output to call openai:\n{result}")
            return result

        else:
            @backoff.on_exception(
                backoff.expo,
                Exception,
                max_tries=12,
                jitter=backoff.random_jitter,
            )
            def fetch():
                global retry_counter
                retry_counter += 1
                token_sub = retry_counter * 200
                try:
                    return openai.ChatCompletion.create(
                            model=model,
                            messages=self.messages_dicts,
                            max_tokens=max_tokens - token_sub,
                            temperature=temperature,
                        ) \
                        .choices[0] \
                        .message["content"]
                except Exception as e:
                    logger.warning(e)
                    raise e
            result = fetch()
            logger.info(f"Output to call openai:\n{result}")
            return result
    
    def call_anthropic(self, model: ChatModel | None = None) -> str:
        if model is None:
            model = self.model
        count_tokens = modal.Function.lookup(UTILS_NAME, "Tiktoken.count")
        messages_length = sum(
            [int(count_tokens.call(message.content) * 1.1) for message in self.messages]
        )
        max_tokens = model_to_max_tokens[model] - int(messages_length) - 1000
        logger.info(f"Number of tokens: {max_tokens}")
        messages_raw = format_for_anthropic(self.messages)
        logger.info(f"Input to call anthropic:\n{messages_raw}")

        assert os.environ.get("ANTHROPIC_API_KEY"), "Please set ANTHROPIC_API_KEY"
        client = anthropic.Client(api_key=os.environ.get("ANTHROPIC_API_KEY"))

        @backoff.on_exception(
            backoff.expo,
            Exception,
            max_tries=12,
            jitter=backoff.random_jitter,
        )
        def fetch() -> tuple[str, str]:
            logger.warning(f"Calling anthropic...")
            results = client.completion(
                prompt=messages_raw,
                stop_sequences=[anthropic.HUMAN_PROMPT],
                model=model,
                max_tokens_to_sample=max_tokens,
                disable_checks=True,
                temperature=temperature,
            )
            return results["completion"], results["stop_reason"]

        result, stop_reason = fetch()
        logger.warning(f"Stop reasons: {stop_reason}")
        if stop_reason == "max_tokens":
            logger.warning("Hit max tokens, running for more tokens.")
            _self = deepcopy(self)
            _self.messages.append(Message(role="assistant", content=result, key=""))
            extension = _self.call_anthropic(model=model)
            print(len(result), len(extension), len(result + extension))
            return result + extension
        logger.info(f"Output to call anthropic:\n{result}")
        return result
    
    def chat_stream(
        self,
        content: str,
        model: ChatModel | None = None,
        message_key: str | None = None,
        functions: list[Function] = [],
        function_call: dict | None = None,
    ) -> Iterator[dict]:
        if self.messages[-1].function_call is None:
            self.messages.append(Message(role="user", content=content, key=message_key))
        else:
            name = self.messages[-1].function_call["name"]
            self.messages.append(Message(role="function", content=content, key=message_key, name=name))
        model = model or self.model
        is_function_call = False
        # might be a bug here in all of this
        # return self.stream_openai(model=model, functions=functions, function_name=function_name)
        return self.stream_openai(model=model, functions=functions, function_call=function_call)
    
    def stream_openai(
        self,
        model: ChatModel | None = None,
        functions: list[Function] = [],
        function_call: dict | None = None,
    ) -> Iterator[dict]:
        model = model or self.model
        count_tokens = modal.Function.lookup(UTILS_NAME, "Tiktoken.count")
        messages_length = sum(
            [count_tokens.call(message.content or "") for message in self.messages]
        )
        max_tokens = model_to_max_tokens[model] - int(messages_length) - 400 # this is for the function tokens
        # TODO: Add a check to see if the message is too long
        logger.info("file_change_paths" + str(self.file_change_paths))
        if len(self.file_change_paths) > 0:
            self.file_change_paths.remove(self.file_change_paths[0])
        if max_tokens < 0:
            if len(self.file_change_paths) > 0:
                pass
            else:
                raise ValueError(f"Message is too long, max tokens is {max_tokens}")
        messages_raw = "\n".join([(message.content or "") for message in self.messages])
        logger.info(f"Input to call openai:\n{messages_raw}")

        gpt_4_buffer = 800
        if int(messages_length) + gpt_4_buffer < 6000 and model == "gpt-4-32k-0613":
            model = "gpt-4-0613"
            max_tokens = model_to_max_tokens[model] - int(messages_length) - gpt_4_buffer # this is for the function tokens

        logger.info(f"Using the model {model}, with {max_tokens} tokens remaining")
        def generator() -> Iterator[str]:
            stream = openai.ChatCompletion.create(
                model=model,
                messages=self.messages_dicts,
                temperature=temperature,
                functions=[json.loads(function.json()) for function in functions],
                function_call=function_call or "auto",
                stream=True
            )
            for data in stream:
                chunk = data.choices[0].delta
                yield chunk
        return generator()

    @property
    def messages_dicts(self):
        # Remove the key from the message object before sending to OpenAI
        cleaned_messages = [
            message.to_openai()
            for message in self.messages
        ]
        return cleaned_messages

    def undo(self):
        if len(self.prev_message_states) > 0:
            self.messages = self.prev_message_states.pop()
        return self.messages



import shutil
import modal
import os
import time
import re
import github
from github import Github
from github.Repository import Repository
from loguru import logger
from git import Repo

from jwt import encode
import requests
from tqdm import tqdm
from sweepai.core.entities import Snippet
from sweepai.utils.config import SweepConfig
from sweepai.utils.constants import APP_ID, DB_NAME
from sweepai.utils.event_logger import posthog


def make_valid_string(string: str):
    pattern = r"[^\w./-]+"
    return re.sub(pattern, "_", string)


def get_jwt():
    signing_key = os.environ["GITHUB_APP_PEM"]
    app_id = APP_ID
    payload = {"iat": int(time.time()), "exp": int(time.time()) + 600, "iss": app_id}

    return encode(payload, signing_key, algorithm="RS256")


def get_token(installation_id: int):
    jwt = get_jwt()
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": "Bearer " + jwt,
        "X-GitHub-Api-Version": "2022-11-28",
    }
    response = requests.post(
        f"https://api.github.com/app/installations/{int(installation_id)}/access_tokens",
        headers=headers,
    )
    return response.json()["token"]


def get_github_client(installation_id: int):
    token = get_token(installation_id)
    return Github(token)


def get_installation_id(username: str):
    jwt = get_jwt()
    response = requests.get(
        f"https://api.github.com/users/{username}/installation",
        headers={
            "Accept": "application/vnd.github+json",
            "Authorization": "Bearer " + jwt,
            "X-GitHub-Api-Version": "2022-11-28",
        },
    )
    obj = response.json()
    try:
        return obj["id"]
    except:
        raise Exception("Could not get installation id, probably not installed")

def display_directory_tree(
    root_path,
    includes: list[str] = [],
    excludes: list[str] = [".git"],
):
    def display_directory_tree_helper(
        current_dir,
        indent="",
    ) -> str:
        files = os.listdir(current_dir)
        files.sort()
        tree = ""
        for item_name in files:
            full_path = os.path.join(current_dir, item_name)[len(root_path) + 1 :]
            if item_name in excludes:
                continue
            file_path = os.path.join(current_dir, item_name)
            if os.path.isdir(file_path):
                if full_path in includes:
                    tree += f"{indent}|- {item_name}/\n"
                    tree += display_directory_tree_helper(
                        file_path, indent + "|   "
                    )
                else:
                    tree += f"{indent}|- {item_name}/...\n"
            else:
                tree += f"{indent}|- {item_name}\n"
        return tree
    tree = display_directory_tree_helper(root_path)
    lines = tree.splitlines()
    return "\n".join([line[3:] for line in lines])


def get_file_list(root_directory: str) -> str:
    files = []

    def dfs_helper(directory):
        nonlocal files
        for item in os.listdir(directory):
            if item == ".git":
                continue
            item_path = os.path.join(directory, item)
            if os.path.isfile(item_path):
                files.append(item_path)  # Add the file to the list
            elif os.path.isdir(item_path):
                dfs_helper(item_path)  # Recursive call to explore subdirectory

    dfs_helper(root_directory)
    files = [file[len(root_directory) + 1 :] for file in files]
    return files


# def get_tree(repo_name: str, installation_id: int) -> str:
#     token = get_token(installation_id)
#     repo_url = f"https://x-access-token:{token}@github.com/{repo_name}.git"
#     Repo.clone_from(repo_url, "repo")
#     tree = display_directory_tree("repo")
#     shutil.rmtree("repo")
#     return tree
def get_tree_and_file_list(
    repo_name: str, 
    installation_id: int, 
    snippet_paths: list[str]
) -> str:
    token = get_token(installation_id)
    shutil.rmtree("repo", ignore_errors=True)
    repo_url = f"https://x-access-token:{token}@github.com/{repo_name}.git"
    Repo.clone_from(repo_url, "repo")

    prefixes = []
    for snippet_path in snippet_paths:
        file_list = ""
        for directory in snippet_path.split("/")[:-1]:
            file_list += directory + "/"
            prefixes.append(file_list.rstrip("/"))
        file_list += snippet_path.split("/")[-1]
        prefixes.append(snippet_path)

    tree = display_directory_tree(
        "repo", 
        includes=prefixes,
    )
    file_list = get_file_list("repo")
    shutil.rmtree("repo")
    return tree, file_list


def get_file_contents(repo: Repository, file_path, ref=None):
    if ref is None:
        ref = repo.default_branch
    file = repo.get_contents(file_path, ref=ref)
    contents = file.decoded_content.decode("utf-8")
    return contents


def search_snippets(
    repo: Repository,
    query: str,
    installation_id: int,
    num_files: int = 5,
    include_tree: bool = True,
    branch: str = None,
    sweep_config: SweepConfig = SweepConfig(),
) -> tuple[Snippet, str]:
    # Initialize the relevant directories string
    get_relevant_snippets = modal.Function.lookup(DB_NAME, "get_relevant_snippets")
    snippets: list[Snippet] = get_relevant_snippets.call(
        repo.full_name, query, num_files, installation_id=installation_id
    )
    logger.info(f"Snippets: {snippets}")
    for snippet in snippets:
        try:
            file_contents = get_file_contents(repo, snippet.file_path, ref=branch)
            if (
                len(file_contents) > sweep_config.max_file_limit
            ):  # more than 10000 tokens
                logger.warning(f"Skipping {snippet.file_path}, too many tokens")
                continue
        except github.UnknownObjectException as e:
            logger.warning(f"Error: {e}")
            logger.warning(f"Skipping {snippet.file_path}")
        else:
            snippet.content = file_contents
    tree, file_list = get_tree_and_file_list(
        repo.full_name, 
        installation_id, 
        snippet_paths=[snippet.file_path for snippet in snippets]
    )
    for file_path in tqdm(file_list):
        if file_path in query:
            try:
                file_contents = get_file_contents(repo, file_path, ref=branch)
                if (
                    len(file_contents) > sweep_config.max_file_limit
                ):  # more than 10000 tokens
                    logger.warning(f"Skipping {file_path}, too many tokens")
                    continue
            except github.UnknownObjectException as e:
                logger.warning(f"Error: {e}")
                logger.warning(f"Skipping {file_path}")
            else:
                snippets.append(
                    Snippet(
                        content=file_contents,
                        start=0,
                        end=file_contents.count("\n") + 1,
                        file_path=file_path,
                    )
                )
    if include_tree:
        return snippets, tree
    else:
        return snippets


def index_full_repository(
    repo_name: str,
    installation_id: int = None,
    sweep_config: SweepConfig = SweepConfig(),
):
    update_index = modal.Function.lookup(DB_NAME, "update_index")
    num_indexed_docs = update_index.spawn(
        repo_name=repo_name,
        installation_id=installation_id,
        sweep_config=sweep_config,
    )
    try:
        repo = get_github_client(installation_id).get_repo(repo_name)
        labels = repo.get_labels()
        label_names = [label.name for label in labels]

        if "sweep" not in label_names:
            repo.create_label(
                name="sweep",
                color="5320E7",
                description="Assigns Sweep to an issue or pull request.",
            )
    except Exception as e:
        posthog("index_full_repository", "failed", {"error": str(e)})
        logger.warning(
            "Adding label failed, probably because label already."
        )  # warn that the repo may already be indexed
    return num_indexed_docs



import os
import requests
from chromadb.utils import embedding_functions

HUGGINGFACE_INFERENCE_URL = "https://r0gyxxwg0w192f22.us-east-1.aws.endpoints.huggingface.cloud"


class HuggingfaceEndpointEmbeddingFunction(embedding_functions.EmbeddingFunction):
    def __init__(self, token: str = os.environ.get("HUGGINGFACE_TOKEN", None)):
        self._token = token
        self._session = requests.Session()
        self._session.headers.update({"Authorization": f"Bearer {token}"})

    def __call__(self, texts):
        # Call HuggingFace Embedding API for each document
        assert self._token
        response = self._session.post(
            HUGGINGFACE_INFERENCE_URL, 
            json={'inputs': texts}
        )
        json = response.json()
        embeddings = json["embeddings"]
        return embeddings




"""
Creates PR given description.
"""

import os
import openai

from loguru import logger
import modal

from sweepai.core.entities import FileChangeRequest, PullRequest
from sweepai.core.sweep_bot import SweepBot
from sweepai.handlers.on_review import review_pr
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import get_github_client
from sweepai.utils.constants import DB_NAME, PREFIX

github_access_token = os.environ.get("GITHUB_TOKEN")
openai.api_key = os.environ.get("OPENAI_API_KEY")

update_index = modal.Function.lookup(DB_NAME, "update_index")

num_of_snippets_to_query = 10
max_num_of_snippets = 5

def create_pr(
    file_change_requests: list[FileChangeRequest],
    pull_request: PullRequest,
    sweep_bot: SweepBot,
    username: str,
    installation_id: int,
    issue_number: int | None = None
):
    # Flow:
    # 1. Get relevant files
    # 2: Get human message
    # 3. Get files to change
    # 4. Get file changes
    # 5. Create PR

    organization, repo_name = sweep_bot.repo.full_name.split("/")
    metadata = {
        "repo_full_name": sweep_bot.repo.full_name,
        "organization": organization,
        "repo_name": repo_name,
        "repo_description": sweep_bot.repo.description,
        "username": username,
        "installation_id": installation_id,
        "function": "on_ticket",
        "mode": PREFIX,
    }
    posthog.capture(username, "started", properties=metadata)

    try:
        logger.info("Making PR...")
        pull_request.branch_name = sweep_bot.create_branch(pull_request.branch_name)
        sweep_bot.change_files_in_github(file_change_requests, pull_request.branch_name)

        # Include issue number in PR description
        if issue_number:
            pr_description = f"{pull_request.content}\n\nFixes #{issue_number}.\n\nTo checkout this PR branch, run the following command in your terminal:\n```zsh\ngit checkout {pull_request.branch_name}\n```"
        else:
            pr_description = f"{pull_request.content}\n\nTo checkout this PR branch, run the following command in your terminal:\n```zsh\ngit checkout {pull_request.branch_name}\n```"

        pr = sweep_bot.repo.create_pull(
            title=pull_request.title,
            body=pr_description,
            head=pull_request.branch_name,
            base=sweep_bot.repo.default_branch,
        )
    except openai.error.InvalidRequestError as e:
        logger.error(e)
        posthog.capture(
            username,
            "failed",
            properties={
                "error": str(e),
                "reason": "Invalid request error / context length",
                **metadata,
            },
        )
        raise e
    except Exception as e:
        logger.error(e)
        posthog.capture(
            username,
            "failed",
            properties={
                "error": str(e),
                "reason": "Unexpected error",
                **metadata,
            },
        )
        raise e

    posthog.capture(username, "success", properties={**metadata})
    logger.info("create_pr success")
    return {"success": True, "pull_request": pr}



import difflib
import re

def generate_diff(old_code, new_code):
    diff = difflib.unified_diff(
        old_code.splitlines(keepends=True),
        new_code.splitlines(keepends=True)
    )
    return ''.join(diff)

def format_contents(file_contents, is_markdown=False):
    """
    Add arbitrary postprocessing here, this affects files and diffs
    """
    lines = file_contents.split('\n')
    code_lines = []
    in_code_block = True

    if is_markdown:
        return '\n'.join(lines) + '\n'
    for line in lines:
        stripped_line = line.strip()

        # Check if line starts a code block
        if stripped_line.startswith('```') and not in_code_block:
            in_code_block = True
            continue

        # Check if line ends a code block
        if stripped_line.endswith('```') and in_code_block:
            in_code_block = False
            continue

        # Append line if it's inside a code block or if it's not an empty line
        if not in_code_block:
            continue
        code_lines.append(line)

    return '\n'.join(code_lines) + '\n'


def generate_new_file(modify_file_response: str, old_file_content: str) -> str:
    import re

    result_file = ""
    old_file_lines = old_file_content.splitlines()

    # Extract content between <new_file> tags
    new_file = re.search(r"<new_file>(.*?)<\/new_file>", modify_file_response, re.DOTALL).group(1).strip()
    if "<copied>" not in new_file:
        return new_file

    # Find all <copied> tags and their content
    copied_sections = re.findall(r"<copied>(.*?)<\/copied>", new_file, re.DOTALL)
    
    first_section_idx = new_file.index("<copied>")
    if first_section_idx > 0:
        result_file += new_file[:first_section_idx]
        new_file = new_file[first_section_idx:] # remove the first section from new_file
    last_section_idx = new_file.rindex("</copied>")
    last_section = ""
    if last_section_idx < len(new_file) - 1:
        last_section = new_file[last_section_idx + len("</copied>"):]
        new_file = new_file[:last_section_idx + len("</copied>")] # remove the last section from new_file
    
    # Parse copied sections, first copying the content and then adding whatever is after the copied section
    for copied_section in copied_sections:
        start_line, end_line = copied_section.split("-")
        start_line = int(start_line) - 1 if int(start_line) - 1 > 0 else 0
        end_line = int(end_line)
        # Check for duplicate lines
        k = 30
        result_file = join_contents_k(result_file, "\n".join(old_file_lines[start_line:end_line]), k)
        new_file = new_file.replace(f"<copied>{copied_section}</copied>\n", "")
        next_section_idx = new_file.index("<copied>") if "<copied>" in new_file else len(new_file)
        # Check for duplicate lines
        result_file = join_contents_k(result_file, new_file[:next_section_idx], k)
        new_file = new_file[next_section_idx:] # remove the first section from new_file
    return result_file + last_section
    
def join_contents_k(first, second, k):
    """
    Join contents together removing k duplicate lines
    """
    first_lines = first.splitlines()
    second_lines = second.splitlines()
    for i in range(k, 0, -1):
        if len(first_lines) < k or len(second_lines) < k:
            continue
        if first_lines[-i:] == second_lines[:i]:
            return "\n".join(first_lines) + "\n" + "\n".join(second_lines[i:])
    return "\n".join(first_lines) + "\n" + "\n".join(second_lines)

def is_markdown(filename):
    return filename.endswith(".md") or filename.endswith(".rst") or filename.endswith(".txt")


import webbrowser
import httpx
from pydantic import BaseModel
import requests
import json
from loguru import logger

from sweepai.app.config import SweepChatConfig
from sweepai.core.entities import PullRequest, Snippet
from sweepai.utils.constants import PREFIX

def break_json(raw_json: str):
    # turns something like {"function_call": {"arguments": " \""}}{"function_call": {"arguments": "summary"}} into two objects
    try:
        yield json.loads(raw_json)
    except json.JSONDecodeError:
        for i in range(1, len(raw_json)):
            try:
                obj = json.loads(raw_json[:i])
                yield obj
                for item in break_json(raw_json[i:]):
                    yield item
                break
            except json.JSONDecodeError:
                pass


class APIClient(BaseModel):
    config: SweepChatConfig
    api_endpoint = f"https://sweepai--{PREFIX}-ui.modal.run"

    def get_installation_id(self):
        results = requests.post(
            self.api_endpoint + "/installation_id",
            json= self.config.dict(),
        )
        if results.status_code == 401:
            print("Installation ID not found! Please install sweep first.")
            webbrowser.open_new_tab("https://github.com/apps/sweep-ai")
            raise Exception(results.json()["detail"])
        if results.status_code != 200:
            raise Exception(results.json()["detail"])
        obj = results.json()
        return obj["installation_id"]

    def search(
        self,
        query: str,
        n_results: int = 5,
    ):
        results = requests.post(
            self.api_endpoint + "/search",
            json={
                "query": query,
                "n_results": n_results,
                "config": self.config.dict(),
            }
        )
        snippets = [Snippet(**item) for item in results.json()]
        return snippets
    
    def create_pr(
        self,
        file_change_requests: list[tuple[str, str]],
        pull_request: PullRequest,
        messages: list[tuple[str | None, str | None]],
    ):
        results = requests.post(
            self.api_endpoint + "/create_pr",
            json={
                "file_change_requests": file_change_requests,
                "pull_request": pull_request,
                "messages": messages,
                "config": self.config.dict(),
            },
            timeout=10 * 60
        )
        return results.json()
    
    def chat(
        self, 
        messages: list[tuple[str | None, str | None]],
        snippets: list[Snippet] = [],
        model: str = "gpt-4-0613",
    ) -> str:
        results = requests.post(
            self.api_endpoint + "/chat",
            json={
                "messages": messages,
                "snippets": [snippet.dict() for snippet in snippets],
                "config": self.config.dict()
            }
        )
        return results.json()
    
    def stream_chat(
        self, 
        messages: list[tuple[str | None, str | None]], 
        snippets: list[Snippet] = [],
        model: str = "gpt-4-0613"
    ):
        print(snippets)
        with httpx.Client(timeout=30) as client: # sometimes this step is slow
            with client.stream(
                'POST', 
                self.api_endpoint + '/chat_stream',
                json={
                    "messages": messages,
                    "snippets": [snippet.dict() for snippet in snippets],
                    "config": self.config.dict()
                }
            ) as response:
                for delta_chunk in response.iter_text():
                    if not delta_chunk:
                        break
                    try:
                        for item in break_json(delta_chunk):
                            yield item
                    except json.decoder.JSONDecodeError as e: 
                        logger.error(delta_chunk)
                        raise e



import os
from loguru import logger
import highlight_io
from posthog import Posthog

POSTHOG_API_KEY = os.environ.get("POSTHOG_API_KEY")
if POSTHOG_API_KEY is None:
    posthog = None
else:
    posthog = Posthog(project_api_key=POSTHOG_API_KEY, host='https://app.posthog.com')
HIGHLIGHT_API_KEY = os.environ.get("HIGHLIGHT_API_KEY")
if HIGHLIGHT_API_KEY is None:
	H = None
else:
	H = highlight_io.H(HIGHLIGHT_API_KEY)
	logger.add(
		H.logging_handler,
		format="{message}",
		level="INFO",
		backtrace=True,
	)



"""
List of common prompts used across the codebase.
"""

# Following two should be fused
system_message_prompt = "Your name is Sweep bot. You are a brilliant and thorough engineer assigned to the following Github ticket. You will be helpful and friendly, but informal and concise: get to the point. When you write code to solve tickets, the code works on the first try and is formatted perfectly. You have the utmost care for the user that you write for, so you do not make mistakes."
system_message_issue_comment_prompt = "Your name is Sweep bot. You are a brilliant and thorough engineer assigned to the following Github ticket, and a user has just responded with feedback. You will be helpful and friendly, but informal and concise: get to the point. When you write code to solve tickets, the code works on the first try and is formatted perfectly. You have the utmost care for the user that you write for, so you do not make mistakes."

human_message_prompt = """
<relevant_snippets_in_repo>
{relevant_snippets}
</relevant_snippets_in_repo>

<relevant_paths_in_repo>
{relevant_directories}
</relevant_paths_in_repo>

<repo_tree>
{tree}
</repo_tree>

Repo: {repo_name}: {repo_description}
Issue Url: {issue_url}
Username: {username}
Issue Title: {title}
Issue Description: {description}
"""

human_message_review_prompt = """
<relevant_snippets_in_repo>
{relevant_snippets}
</relevant_snippets_in_repo>

<relevant_paths_in_repo>
{relevant_directories}
</relevant_paths_in_repo>

<repo_tree>
{tree}
</repo_tree>

These are the file changes.
We have the file_path, the previous_file_content, the new_file_content, and the diffs.
The file_path is the name of the file.
The previous_file_content is the content of the file before the changes.
The new_file_content is the content of the file after the changes.
The diffs are the lines changed in the file. <added_lines> indicates those lines were added, <deleted_lines> indicates they were deleted.
Keep in mind that we may see a diff for a deletion and replacement, so don't point those out as issues.
{diffs}
"""

diff_section_prompt = """
<file_path>
{diff_file_path}
</file_path>

<previous_file_content>
{previous_file_content}
</previous_file_content>

<new_file_content>
{new_file_content}
</new_file_content>

<file_diffs>
{diffs}
</file_diffs>
"""

review_prompt = """\
I need you to carefully review the code diffs in this pull request. 
The code was written by an inexperienced programmer and may contain accidental deletions, logic errors or other issues.
Think step-by-step logically and thoroughly analyze to summarize the diffs per file in the format:

Step-by-step thoughts:
* Lines x1-x2: Summary of the changes (added, deleted, modified, errors, issues) 
* Lines y1-y2: Summary of the changes (added, deleted, modified, errors, issues)
...
<file_summarization>
* file_1 - changes in file_1
* file_1 - more changes in file_1
...
</file_summarization>
"""

review_follow_up_prompt = """\
Here is the next file diff.
Think step-by-step logically and accurately to summarize the diffs per file in the format:
Step-by-step thoughts:
* Lines x1-x2: Summary of the changes (added, deleted, modified, errors, issues) 
* Lines y1-y2: Summary of the changes (added, deleted, modified, errors, issues)
...
<file_summarization>
* file_1 - changes in file_1
* file_1 - more changes in file_1
...
</file_summarization>
"""

final_review_prompt = """\
This were the file summaries you provided:
<file_summaries>
{file_summaries}
</file_summaries>
Given these summaries write a direct and concise GitHub review comment. If there are no changes required, simply say "No changes required."
In case changes are required, keep in mind the author is an inexperienced programmer and may need a pointer to the files and specific changes.
Follow this format:
<changes_required>
Write Yes if the changes are required or No if they are not required.
</changes_required>
<review_comment>
Mention any changes that need to be made, using GitHub markdown to format the comment.
- Change required in file on line x1-x2
- Change required in file on line y1-y2
...
</review_comment>
"""

issue_comment_prompt = """
<comment username="{username}">
{reply}
</comment>
"""

# Prompt for comments
human_message_prompt_comment = """
<relevant_snippets_in_repo>
{relevant_snippets}
</relevant_snippets_in_repo>

<relevant_paths_in_repo>
{relevant_directories}
</relevant_paths_in_repo>

<repo_tree>
{tree}
</repo_tree>

Repo: {repo_name}: {repo_description}
Issue Url: {issue_url}
Username: {username}
Pull Request Title: {title}
Pull Request Description: {description}

These are the file changes.
We have the file_path, the previous_file_content, the new_file_content, and the diffs.
The file_path is the name of the file.
The previous_file_content is the content of the file before the changes.
The new_file_content is the content of the file after the changes.
The diffs are the lines changed in the file. <added_lines> indicates those lines were added, <deleted_lines> indicates they were deleted.
Keep in mind that we may see a diff for a deletion and replacement, so don't point those out as issues.
{diff}
Please handle the user review comment, taking into account the snippets, paths, tree, pull request title, pull request description, and the file changes.
Sometimes the user may not request changes, don't change anything in that case.
User pull request review: {comment}
"""

comment_line_prompt = """\
The user made the review in this file: {pr_file_path}
and on this line: {pr_line}
"""

cot_retrieval_prompt = """
Gather information to solve the problem. Use "finish" when you feel like you have sufficient information.
""" 

files_to_change_prompt = """
Think step-by-step to break down the requested problem or feature, and then figure out what to change in the current codebase.
Then, provide a list of files you would like to modify, abiding by the following:
* Including the FULL path, e.g. src/main.py and not just main.py
* Use a one-line, detailed, natural language instructions on what to modify, with reference to variable names
* The list of files to create or modify may be empty, but you MUST leave the XML tags with a single list element with "* None"
* There MUST be both create and modify XML tags
* Create/modify up to 5 FILES
* Do not modify non-text files, like images, svgs, binary, etc
* You MUST follow the following format:

Step-by-step thoughts with explanations: 
* Thought 1 - Explanation 1
* Thought 2 - Explanation 2
...

<create>
* filename_1: instructions_1
* filename_2: instructions_2
...
</create>

<modify>
* filename_3: instructions_3
* filename_4: instructions_4
...
</modify>
"""

reply_prompt = """
Write a response to this user:
* Ping the user.
* Tell them you have started working on this PR and a rough summary of your plan. 
* Do not start with "Here is a draft", just write the response.
* End with "Give me a minute!".
* Use github markdown to format the response.
"""

create_file_prompt = """
Think step-by-step regarding the instructions and what should be added to the new file.
Then create a plan of parts of the code to create, with low-level, detailed references to functions and variable to create, and what each function does.
Then create the following file using the following instructions:

File Name: {filename}

Instructions: {instructions}

Reply in the following format. DO NOT write "pass" or "Rest of code". Do not literally write "{{new_file}}". You must use the new_file XML tags, and all text inside these tags will be placed in the newly created file.

Step-by-step thoughts with explanations: 
* Thought 1 - Explanation 1
* Thought 2 - Explanation 2
...
Detailed plan of additions:
* Addition 1
* Addition 2
...
Commit Message: {{commit_message}}
<new_file>
{{new_file}}
</new_file>
"""

"""
Reply in the format below. 
* You MUST use the new_file XML tags
* DO NOT write ``` anywhere, unless it's markdown
* DO NOT write "pass" or "Rest of code"
* Do not literally write "{{new_file}}".
* Format:
"""

modify_file_plan_prompt = """
Think step-by-step regarding the instructions and how that can be applied to the current file to improve the current codebase.
Then create a plan of parts of the code to modify with detailed references to functions to modify.

File Name: {filename}
<old_file>
{code}
</old_file>

Your instructions to modify the file are: "{instructions}".

Step-by-step thoughts with explanations: 
* Thought 1 - Explanation 1
* Thought 2 - Explanation 2
...
Detailed plan of modifications:
* Modification 1
* Modification 2
...
"""

modify_file_prompt = """
Generate a new_file based on the given plan, ensuring that you:
1. Do not write "pass" statements.
2. Provide complete functions with actual business logic. It is imperative that we do not leave any work to the user/future readers of this code.
3. Do not write new "todo" comments.
4. Do not write incomplete functions or line numbers.
5. Make sure code follows programming language conventions in repo

Instead of writing "# Rest of Code", specify the lines to copy from the old file using an XML tag, inclusive (e.g., "<copied>0-25</copied>"). Make sure to use this exact format.
Copy the correct line numbers and copy as long of a prefix and suffix as possible. For instance, if you want to insert code after line 50, start with "<copied>0-50</copied>".

Example: New file:
print("new file")
</new_file>

Example: Insert at end:
<copied>0-100</copied>
print("inserted at end")
</new_file>

Example: If you want to insert code after lines 50 and 75:
<new_file>
<copied>0-50</copied>
def main():
     print("hello world")
<copied>51-75</copied>
print("debug statement")
<copied>76-100</copied>
</new_file>
"""

pr_code_prompt = ""  # TODO: deprecate this


pull_request_prompt = """
Awesome! Could you also provide a PR message in the following format? Content should be in Github style markdown. Thanks!

Title: {title}
Branch Name: {branch_name}
<content>
{content}
</content>
"""

summarize_system_prompt = """
Your name is Sweep bot. You are an engineer assigned to helping summarize code instructions and code changes.
"""

user_file_change_summarize_prompt = """
Summarize the given instructions for making changes in a pull request.
Code Instructions:
{message_content}
"""

assistant_file_change_summarize_prompt = """
Please summarize the following file using the file stubs. 
Be sure to repeat each method signature and docstring. You may also add additional comments to the docstring. 
Do not repeat the code in the file stubs.
Code Changes:
{message_content}
"""

slack_system_message_prompt = "Your name is Sweep bot. You are an engineer assigned to assisting the following Slack user. You will be helpful and friendly, but informal and concise: get to the point. You will use Slack-style markdown when needed to structure your responses."

slack_slash_command_prompt = """
Relevant snippets provided by search engine (decreasing relevance):
<relevant_snippets_in_repo>
{relevant_snippets}
</relevant_snippets_in_repo>

<relevant_paths_in_repo>
{relevant_directories}
</relevant_paths_in_repo>

Repo: {repo_name}: {repo_description}
Username: {username}
Query: {query}

Gather information (i.e. fetch more snippets) to solve the problem. Use "create_pr" if the user asks for changes or you think code changes are needed.
"""

code_repair_system_prompt = """\
You are a genius trained for code repair. 
You will be given two pieces of code marked by xml tags. The code inside <diff></diff> is the difference betwen the user_code and the original code, and the code inside <user_code></user_code> is a user's attempt at adding a change described as {feature}. 
Our goal is to return a working version of user_code that follows {feature}.

Instructions:
* Keep the logic changes from user_code.
* Fix any issues using our knowledge of both the diff and user_code files. 
* Fix syntax errors and accidentally deleted lines.
* Do not perform code style cleanup.
* Do not add or remove any whitespace besides what is necessary to fix syntax errors.
* Do not add or remove any comments.
"""

code_repair_prompt = """\
<diff>
{diff}
</diff>
<user_code>
{user_code}
</user_code>
This is the user_code. 
Return the repaired user_code without xml tags. All of the text you return will be placed in the file. Revert any unrelated deletions to user_code, using the diff and described change.
"""

gradio_system_message_prompt = """Your name is Sweep bot. You are a brilliant and thorough engineer assigned to assist the following user with their problems in the Github repo. You will be helpful and friendly, but informal and concise: get to the point. When you write code to solve tickets, the code works on the first try and is formatted perfectly. You have the utmost care for the user that you write for, so you do not make mistakes. If the user asks you to create a PR, you will use the create_pr function.

Relevant snippets provided by search engine (decreasing relevance):
{snippets}
Repo: {repo_name}
Description: {repo_description}"""


"""
On Github ticket, get ChatGPT to deal with it
"""

# TODO: Add file validation

import os
import openai

from loguru import logger
import modal

from sweepai.core.entities import FileChangeRequest, Snippet
from sweepai.core.prompts import (
    reply_prompt,
)
from sweepai.core.sweep_bot import SweepBot
from sweepai.core.prompts import issue_comment_prompt
from sweepai.handlers.create_pr import create_pr
from sweepai.handlers.on_comment import on_comment
from sweepai.handlers.on_review import review_pr
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import get_github_client, search_snippets
from sweepai.utils.prompt_constructor import HumanMessagePrompt
from sweepai.utils.constants import DB_NAME, PREFIX, UTILS_NAME

github_access_token = os.environ.get("GITHUB_TOKEN")
openai.api_key = os.environ.get("OPENAI_API_KEY")

update_index = modal.Function.lookup(DB_NAME, "update_index")

bot_suffix = "I'm a bot that handles simple bugs and feature requests \
but I might make mistakes. Please be kind!"

collapsible_template = """
<details>
  <summary>{summary}</summary>

  {body}
</details>
"""

chunker = modal.Function.lookup(UTILS_NAME, "Chunking.chunk")

num_of_snippets_to_query = 30
max_num_of_snippets = 5

def on_ticket(
    title: str,
    summary: str,
    issue_number: int,
    issue_url: str,
    username: str,
    repo_full_name: str,
    repo_description: str,
    installation_id: int,
    comment_id: int = None
):
    # Check if the title starts with "sweep" or "sweep: " and remove it
    if title.lower().startswith("sweep: "):
        title = title[7:]
    elif title.lower().startswith("sweep "):
        title = title[6:]

    # Flow:
    # 1. Get relevant files
    # 2: Get human message
    # 3. Get files to change
    # 4. Get file changes
    # 5. Create PR

    organization, repo_name = repo_full_name.split("/")
    metadata = {
        "issue_url": issue_url,
        "issue_number": issue_number,
        "repo_full_name": repo_full_name,
        "organization": organization,
        "repo_name": repo_name,
        "repo_description": repo_description,
        "username": username,
        "installation_id": installation_id,
        "function": "on_ticket",
        "mode": PREFIX,
    }
    posthog.capture(username, "started", properties=metadata)

    g = get_github_client(installation_id)

    if comment_id:
        logger.info(f"Replying to comment {comment_id}...")
    logger.info(f"Getting repo {repo_full_name}")
    repo = g.get_repo(repo_full_name)
    current_issue = repo.get_issue(number=issue_number)
    if current_issue.state == 'closed':
        posthog.capture(username, "issue_closed", properties=metadata)
        return {"success": False, "reason": "Issue is closed"}
    item_to_react_to = current_issue.get_comment(comment_id) if comment_id else current_issue
    eyes_reaction = item_to_react_to.create_reaction("eyes")

    def comment_reply(message: str):
        current_issue.create_comment(message + "\n\n---\n" + bot_suffix)

    comments = current_issue.get_comments()
    replies_text = ""
    if comment_id:
        replies_text = "\nComments:\n" + "\n".join(
            [
                issue_comment_prompt.format(
                    username=comment.user.login,
                    reply=comment.body,
                ) for comment in comments
            ]
        )

    def fetch_file_contents_with_retry():
        retries = 3
        error = None
        for i in range(retries):
            try:
                logger.info(f"Fetching relevant files for the {i}th time...")
                return search_snippets(
                    repo,
                    f"{title}\n{summary}\n{replies_text}",
                    num_files=num_of_snippets_to_query,
                    branch=None,
                    installation_id=installation_id,
                )
            except Exception as e:
                error = e
                continue
        posthog.capture(
            username, "fetching_failed", properties={"error": error, **metadata}
        )
        raise error

    logger.info("Fetching relevant files...")
    try:
        snippets, tree = fetch_file_contents_with_retry()
        assert len(snippets) > 0
    except Exception as e:
        logger.error(e)
        comment_reply(
            "It looks like an issue has occured around fetching the files. Perhaps the repo has not been initialized: try removing this repo and adding it back. I'll try again in a minute. If this error persists contact team@sweep.dev."
        )
        raise e

    num_full_files = 2
    num_extended_snippets = 2

    most_relevant_snippets = snippets[:num_full_files]
    snippets = snippets[:-num_full_files]
    logger.info("Expanding snippets...")
    for snippet in most_relevant_snippets:
        current_snippet = snippet
        _chunks, metadatas, _ids = chunker.call(
            current_snippet.content, 
            current_snippet.file_path
        )
        segmented_snippets = [
            Snippet(
                content=current_snippet.content,
                start=metadata["start"],
                end=metadata["end"],
                file_path=metadata["file_path"],
            ) for metadata in metadatas
        ]
        index = 0
        while index < len(segmented_snippets) and segmented_snippets[index].start <= current_snippet.start:
            index += 1
        index -= 1
        for i in range(index + 1, min(index + num_extended_snippets + 1, len(segmented_snippets))):
            current_snippet += segmented_snippets[i]
        for i in range(index - 1, max(index - num_extended_snippets - 1, 0), -1):
            current_snippet = segmented_snippets[i] + current_snippet
        snippets.append(current_snippet)

    # snippet fusing
    i = 0
    while i < len(snippets):
        j = i + 1
        while j < len(snippets):
            if snippets[i] ^ snippets[j]:  # this checks for overlap
                snippets[i] = snippets[i] | snippets[j]  # merging
                snippets.pop(j)
            else:
                j += 1
        i += 1

    snippets = snippets[:min(len(snippets), max_num_of_snippets)]

    human_message = HumanMessagePrompt(
        repo_name=repo_name,
        issue_url=issue_url,
        username=username,
        repo_description=repo_description,
        title=title,
        summary=summary + replies_text,
        snippets=snippets,
        tree=tree, # TODO: Anything in repo tree that has something going through is expanded
    )
    sweep_bot = SweepBot.from_system_message_content(
        human_message=human_message, repo=repo, is_reply=bool(comments)
    )
    sweepbot_retries = 3
    try:
        for i in range(sweepbot_retries):
            logger.info("CoT retrieval...")
            if sweep_bot.model == "gpt-4-32k-0613":
                sweep_bot.cot_retrieval()
            logger.info("Fetching files to modify/create...")
            file_change_requests = sweep_bot.get_files_to_change()
            for file_change_request in file_change_requests:
                try:
                    contents = repo.get_contents(file_change_request.filename)
                    if contents:
                        file_change_request.change_type = "modify"
                    else:
                        file_change_request.change_type = "create"
                except:
                    file_change_request.change_type = "create"
            logger.info("Getting response from ChatGPT...")
            reply = sweep_bot.chat(reply_prompt, message_key="reply")
            sweep_bot.delete_messages_from_chat("reply")
            logger.info("Sending response...")
            new_line = '\n'
            comment_reply(
                reply
                + "\n\n"
                + collapsible_template.format(
                    summary="Some code snippets I looked at (click to expand). If some file is missing from here, you can mention the path in the ticket description.",
                    body="\n".join(
                        [
                            f"https://github.com/{organization}/{repo_name}/blob/{repo.get_commits()[0].sha}/{snippet.file_path}#L{max(snippet.start, 1)}-L{min(snippet.end, snippet.content.count(new_line))}\n"
                            for snippet in snippets
                        ]
                    ),
                )
            )

            logger.info("Generating PR...")
            pull_request = sweep_bot.generate_pull_request()

            logger.info("Making PR...")
            response = create_pr(file_change_requests, pull_request, sweep_bot, username, installation_id, issue_number)
            if not response or not response["success"]: raise Exception("Failed to create PR")
            pr = response["pull_request"]
            current_issue.create_reaction("rocket")
            try:
                eyes_reaction.delete()
            except:
                pass
            try:
                changes_required, review_comment = review_pr(repo=repo, pr=pr, issue_url=issue_url, username=username, 
                        repo_description=repo_description, title=title, 
                        summary=summary, replies_text=replies_text, tree=tree)
                logger.info(f"Addressing review comment {review_comment}")
                if changes_required:
                    on_comment(repo_full_name=repo_full_name, 
                            repo_description=repo_description, 
                            comment=review_comment,
                            username=username, 
                            installation_id=installation_id,
                            pr_path=None,
                            pr_line_position=None,
                            pr_number=pr.number)
            except Exception as e:
                logger.error(e)
            break
    except openai.error.InvalidRequestError as e:
        logger.error(e)
        comment_reply(
            "I'm sorry, but it looks our model has ran out of context length. We're trying to make this happen less, but one way to mitigate this is to code smaller files. If this error persists contact team@sweep.dev."
        )
        posthog.capture(
            username,
            "failed",
            properties={
                "error": str(e),
                "reason": "Invalid request error / context length",
                **metadata,
            },
        )
        raise e
    except Exception as e:
        logger.error(e)
        comment_reply(
            "I'm sorry, but it looks like an error has occured. Try removing and re-adding the sweep label. If this error persists contact team@sweep.dev."
        )
        posthog.capture(
            username,
            "failed",
            properties={"error": str(e), "reason": "Generic error", **metadata},
        )
        raise e
    else:
        try:
            eyes_reaction.delete()
        except:
            pass
        item_to_react_to.create_reaction("rocket")

    posthog.capture(username, "success", properties={**metadata})
    logger.info("on_ticket success")
    return {"success": True}




from datetime import datetime

def compute_score(contents, commits):
    line_count = contents.count("\n") 
    if line_count > 30:
        line_count_score = 2
    else:
        line_count_score = line_count / 15
    commit_count = len(commits)
    commit_count = 1 if commit_count < 3 else commit_count / 3
    days_since_last_modified = (datetime.now() - commits[0].commit.author.date).days + 1
    if days_since_last_modified > 14:
        days_since_last_modified_score = 7 / days_since_last_modified
    elif days_since_last_modified <= 4:
        days_since_last_modified_score = 2
    else:
        days_since_last_modified_score = 10 / days_since_last_modified
    return (line_count_score * commit_count * days_since_last_modified_score)


"""
Utils for react agent
"""

import re
from textwrap import dedent
from typing import Callable, Concatenate

from pydantic import BaseModel

REACT_INITIAL_PROMPT = """
Gather information to solve the above problem using the tools below. 
* You should use the tools about 2-3 times
* Only use Finish when you are CERTAIN you have enough information to solve the problem. More information is usually better.
* The examples are provided ONLY as examples: the example inputs probably will not work
* You are given the following tools:

{tools}

Respond in the following format, first thinking logically and deriving a plan, then respond with the tool, and inputs replaced by their values:
Thoughts:
1. thought_1
2. thought_2
...
Plan: 
1. plan_1
2. plan_2
...
<tool>tool</tool>
<inputs>
inputs
</inputs>
"""

REACT_RESPONSE_PROMPT = """
The tool returned the following output.

<output>
{output}
</output>

What would you like to do next? Use the same format as above.
"""

def dedent(text: str) -> str:
    return re.sub(r'(\n)[ \t]+', r'\n', text.strip())

class Tool(BaseModel):
    description: str
    function: Callable[[str], str]
    example_inputs: str  = ""
    name: str = ""

    @property
    def _name(self):
        return self.__class__.__name__ if self.name == "" else self.name
    
    @property
    def summary(self):
        return dedent(f"""
        {self._name}: {self.description}
        Example usage:
        <tool>{self._name}</tool>
        <inputs>
        {self.example_inputs}
        </inputs>
        """.strip())
    
    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)
    
    @classmethod
    def tool(cls, **kwargs):
        def decorator(function):
            return cls(function=function, **kwargs)
        return decorator

class CodeSearch(Tool):
    description = "Search in the codebase for relevant snippets of code. Takes natural language search query as input."
    example_inputs = "Modal component on main page."

class ReadFiles(Tool):
    description = "Reads listed files. Takes list of literal file paths as input, separated by newlines. Max 3 files." 
    example_inputs = "src/main.py\ntests/test_main.py\ntests/test_utils.py:101:200"

class Google(Tool):
    description = "Search for code documentation on Google. Takes natural language search query as input."
    example_inputs = "Discord API docs"

class Finish(Tool):
    description = "Indicate you have sufficient information to move forward with making changes to the codebase. Return with an empty string."
    example_inputs = ""
    function = lambda _: ""

class Toolbox(BaseModel):
    tools: list[Tool] = []

    @property
    def prompt(self):
        return REACT_INITIAL_PROMPT.format(tools="\n\n".join([tool.summary for tool in self.tools]))

    class ParsedResults(BaseModel):
        tool_name: str
        inputs: str
        _regex: str = r"<tool>(?P<tool>.*?)</tool>\s+<inputs>(?P<inputs>.*?)</inputs>"

        @classmethod
        def parse(cls, results) -> "Toolbox.ParsedResults":
            match = re.search(cls._regex, results, flags=re.DOTALL)
            return cls(tool_name=match.group("tool").strip(), inputs=match.group("inputs").strip())

    def process_results(self, parsed_results: "Toolbox.ParsedResults") -> str:
        # parsed_results = Toolbox.ParsedResults.parse(raw_output)
        tool = next((tool for tool in self.tools if tool._name == parsed_results.tool_name), None)
        return tool(parsed_results.inputs)




from pydantic import BaseModel

class SweepConfig(BaseModel):
    include_dirs: list[str] = []
    exclude_dirs: list[str] = [".git", "node_modules", "venv"]
    include_exts: list[str] = ['.cs', '.csharp', '.py', '.md', '.txt', '.ts', '.tsx', '.js', '.jsx', '.mjs']
    exclude_exts: list[str] = ['.min.js', '.min.js.map', '.min.css', '.min.css.map']
    max_file_limit: int = 60_000



"""
This file should be split into environment and config files
"""

PREFIX = "dev"
DB_NAME = PREFIX + "-db"
API_NAME = PREFIX + "-api"
UTILS_NAME = PREFIX + "-utils"
SLACK_NAME = PREFIX + "-slack"
BOT_TOKEN_NAME = PREFIX + "-bot-token"
if PREFIX == "prod":
    BOT_TOKEN_NAME = "bot-token"
SWEEP_LOGIN = "sweep-ai[bot]"

if PREFIX == "prod":
    APP_ID = 307814
    ENV = PREFIX
elif PREFIX == "dev2":
    APP_ID = 327588
    ENV = PREFIX
    SWEEP_LOGIN = "sweep-canary[bot]"
elif PREFIX == "dev":
    APP_ID = 324098
    ENV = PREFIX
    SWEEP_LOGIN = "sweep-nightly[bot]"
LABEL_NAME = "sweep"
LABEL_COLOR = "#9400D3"
LABEL_DESCRIPTION = "Sweep your software chores"


"""
On Github ticket, get ChatGPT to deal with it
"""

# TODO: Add file validation

import os
import openai

from loguru import logger

from sweepai.core.sweep_bot import SweepBot
from sweepai.handlers.on_review import get_pr_diffs
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import (
    get_github_client,
    search_snippets,
)
from sweepai.utils.prompt_constructor import HumanMessageCommentPrompt
from sweepai.utils.constants import PREFIX

github_access_token = os.environ.get("GITHUB_TOKEN")
openai.api_key = os.environ.get("OPENAI_API_KEY")


def on_comment(
    repo_full_name: str,
    repo_description: str,
    comment: str,
    pr_path: str | None,
    pr_line_position: int | None,
    username: str,
    installation_id: int,
    pr_number: int = None,
):
    # Check if the comment is "REVERT"
    if comment.strip().upper() == "REVERT":
        rollback_file(repo_full_name, pr_path, installation_id, pr_number)
        return {"success": True, "message": "File has been reverted to the previous commit."}

    # Flow:
    # 1. Get relevant files
    # 2: Get human message
    # 3. Get files to change
    # 4. Get file changes
    # 5. Create PR
    logger.info(f"Calling on_comment() with the following arguments: {comment}, {repo_full_name}, {repo_description}, {pr_path}")
    organization, repo_name = repo_full_name.split("/")
    metadata = {
        "repo_full_name": repo_full_name,
        "repo_name": repo_name,
        "organization": organization,
        "repo_description": repo_description,
        "installation_id": installation_id,
        "username": username,
        "function": "on_comment",
        "mode": PREFIX,
    }

    posthog.capture(username, "started", properties=metadata)
    logger.info(f"Getting repo {repo_full_name}")
    try:
        g = get_github_client(installation_id)
        repo = g.get_repo(repo_full_name)
        pr = repo.get_pull(pr_number)
        # Check if the PR is closed
        if pr.state == "closed":
            return {"success": True, "message": "PR is closed. No event fired."}
        branch_name = pr.head.ref
        pr_title = pr.title
        pr_body = pr.body
        diffs = get_pr_diffs(repo, pr)
        snippets, tree = search_snippets(repo, comment, installation_id, branch=branch_name, num_files=1 if pr_path else 3)
        pr_line = None
        pr_file_path = None
        if pr_path and pr_line_position:
            pr_file = repo.get_contents(pr_path, ref=branch_name).decoded_content.decode("utf-8")
            pr_lines = pr_file.splitlines()
            pr_line = pr_lines[min(len(pr_lines), pr_line_position) - 1]
            pr_file_path = pr_path.strip()

        logger.info("Getting response from ChatGPT...")
        human_message = HumanMessageCommentPrompt(
            comment=comment,
            repo_name=repo_name,
            repo_description=repo_description if repo_description else "",
            diffs=diffs,
            issue_url=pr.html_url,
            username=username,
            title=pr_title,
            tree=tree,
            summary=pr_body,
            snippets=snippets,
            pr_file_path=pr_file_path, # may be None
            pr_line=pr_line, # may be None
        )
        logger.info(f"Human prompt{human_message.construct_prompt()}")
        sweep_bot = SweepBot.from_system_message_content(
            # human_message=human_message, model="claude-v1.3-100k", repo=repo
            human_message=human_message, repo=repo, 
        )
    except Exception as e:
        posthog.capture(username, "failed", properties={
            "error": str(e),
            "reason": "Failed to get files",
            **metadata
        })
        raise e

    try:
        logger.info("Fetching files to modify/create...")
        file_change_requests = sweep_bot.get_files_to_change()

        logger.info("Making Code Changes...")
        sweep_bot.change_files_in_github(file_change_requests, branch_name)

        logger.info("Done!")
    except Exception as e:
        posthog.capture(username, "failed", properties={
            "error": str(e),
            "reason": "Failed to make changes",
            **metadata
        })
        raise e

    posthog.capture(username, "success", properties={**metadata})
    logger.info("on_comment success")
    return {"success": True}

def rollback_file(repo_full_name, pr_path, installation_id, pr_number):
    g = get_github_client(installation_id)
    repo = g.get_repo(repo_full_name)
    pr = repo.get_pull(pr_number)
    branch_name = pr.head.ref

    # Get the file's content from the previous commit
    commits = repo.get_commits(sha=branch_name)
    if commits.totalCount < 2:
        current_file = repo.get_contents(pr_path, ref=commits[0].sha)
        current_file_sha = current_file.sha
        previous_content = repo.get_contents(pr_path, ref=repo.default_branch)
        previous_file_content = previous_content.decoded_content.decode("utf-8")
        repo.update_file(pr_path, "Revert file to previous commit", previous_file_content, current_file_sha, branch=branch_name)
        return
    previous_commit = commits[1]
    
    # Get current file SHA
    current_file = repo.get_contents(pr_path, ref=commits[0].sha)
    current_file_sha = current_file.sha

    # Check if the file exists in the previous commit
    try:
        previous_content = repo.get_contents(pr_path, ref=previous_commit.sha)
        previous_file_content = previous_content.decoded_content.decode("utf-8")
        # Create a new commit with the previous file content
        repo.update_file(pr_path, "Revert file to previous commit", previous_file_content, current_file_sha, branch=branch_name)
    except Exception as e:
        if e.status == 404:
            logger.warning(f"File {pr_path} was not found in previous commit {previous_commit.sha}")
        else:
            raise e




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sweepai', 'sweepai.app', 'sweepai.core', 'sweepai.handlers', 'sweepai.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyGithub==1.58.2',
 'config-path>=1.0.3,<2.0.0',
 'fastapi>=0.94.1,<0.95.0',
 'gradio>=3.35.2,<4.0.0',
 'loguru>=0.6.0,<0.7.0',
 'requests>=2.28.2,<3.0.0',
 'typer>=0.9.0,<0.10.0',
 'urllib3>=2.0.3,<3.0.0',
 'uvicorn>=0.21.0,<0.22.0']

entry_points = \
{'console_scripts': ['sweep = sweepai.app.cli:app',
                     'sweepai = sweepai.app.cli:app']}

setup_kwargs = {
    'name': 'sweepai',
    'version': '0.1.10',
    'description': 'Sweep software chores',
    'long_description': None,
    'author': 'Kevin Lu',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



import json
import os

from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.responses import RedirectResponse

import modal
from pydantic import BaseModel
from pymongo import MongoClient
import requests
from sweepai.core.entities import FileChangeRequest, Function, PullRequest
import slack_sdk
from loguru import logger
from slack_sdk.oauth.installation_store import Installation, InstallationStore, Bot

from sweepai.core.sweep_bot import SweepBot
from sweepai.utils.github_utils import get_github_client
from sweepai.utils.constants import API_NAME, BOT_TOKEN_NAME, PREFIX, SLACK_NAME
from sweepai.core.prompts import slack_slash_command_prompt
from sweepai.utils.github_utils import get_installation_id
import sweepai.utils.event_logger

image = modal.Image \
    .debian_slim() \
    .apt_install("git") \
    .pip_install(
        "slack-sdk",
        "slack-bolt",
        "pymongo",
        "PyGithub",
        "gitpython",
        "openai",
        "anthropic",
        "loguru",
        "tqdm",
        "highlight-io",
        "posthog",
        "pyyaml"
    )
stub = modal.Stub(SLACK_NAME)
secrets = [
    modal.Secret.from_name("slack"),
    modal.Secret.from_name("openai-secret"),
    modal.Secret.from_name(BOT_TOKEN_NAME),
    modal.Secret.from_name("mongodb"),
]

functions = [
    Function(
        name="create_pr",
        description="Creates a PR.",
        parameters={
            "properties": {
                "plan": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "file_path": {
                                "type": "string",
                                "description": "The file path to change."
                            },
                            "instructions": {
                                "type": "string",
                                "description": "Detailed description of what the change as a list."
                            },
                        },
                        "required": ["file_path", "instructions"]
                    },
                    "description": "A list of files to modify or create and instructions for what to modify or create."
                },
                "title": {
                    "type": "string",
                    "description": "Title of PR",
                },
                "summary": {
                    "type": "string",
                    "description": "Detailed summary of PR",
                },
                "branch": {
                    "type": "string",
                    "description": "Name of branch to create PR in.",
                },
            }
        }
    ),
    Function(
        name="get_relevant_snippets",
        description="Search engine for relevant snippets in the current repo in natural language.",
        parameters={
            "properties": {
                "query": {
                    "type": "string",
                    "description": "Natural language query to search."
                }
            }
        }
    )
]

pr_format = """I'm going to create a PR with the following:

> *{title}*
> {summary}

I have the following plan:
{plan}

:hourglass_flowing_sand: Creating...
"""

pr_done_format = """:white_check_mark: Done creating PR at {url} with the following:
> *{title}*
> {summary}
"""

thread_query_format = "Message thread: {messages}"

# TODO(sweep): move to constants
slack_app_page = "https://sweepusers.slack.com/apps/A05D69L28HX-sweep"
slack_install_link = "https://slack.com/oauth/v2/authorize?client_id=5364586338420.5448326076609&scope=channels:read,chat:write,chat:write.public,commands,groups:read,im:read,incoming-webhook,mpim:read&user_scope="

class MongoDBInstallationStore(InstallationStore):
    def __init__(self):
        self.client = MongoClient(os.environ['MONGODB_URI'])
        self.db = self.client["slack"]
        self.installation_collection = self.db["installation"]
        self.bot_collection = self.db["bot"]
    
    def save(self, installation: Installation):
        self.installation_collection.replace_one({
            "user_id": installation.user_id,
            "team_id": installation.team_id,
            "enterprise_id": installation.enterprise_id,
            "is_enterprise_install": installation.is_enterprise_install,
        }, {
            "installation": installation.to_dict(),
            "user_id": installation.user_id,
            "team_id": installation.team_id,
            "enterprise_id": installation.enterprise_id,
            "is_enterprise_install": installation.is_enterprise_install,
            "prefix": PREFIX,
            "access_token": installation.bot_token,
        }, upsert=True)

    def save_bot(self, bot: Bot):
        self.bot_collection.replace_one({
            "team_id": bot.team_id,
            "enterprise_id": bot.enterprise_id,
            "is_enterprise_install": bot.is_enterprise_install,
        }, {
            "installation": bot.to_dict(),
            "team_id": bot.team_id,
            "enterprise_id": bot.enterprise_id,
            "is_enterprise_install": bot.is_enterprise_install,
            "prefix": PREFIX,
            "access_token": bot.bot_token,
        }, upsert=True)
    
    def find_installation(
        self,
        *,
        enterprise_id: str,
        team_id: str,
        is_enterprise_install: bool,
        user_id: str | None = None,
    ):
        if user_id:
            return Installation(**self.installation_collection.find_one({
                "user_id": user_id,
                "team_id": team_id,
                "enterprise_id": enterprise_id,
                "is_enterprise_install": is_enterprise_install
            })["installation"])
        else:
            return Installation(**self.installation_collection.find_one({
                "team_id": team_id,
                "enterprise_id": enterprise_id,
                "is_enterprise_install": is_enterprise_install
            })["installation"])

    def find_bot(
        self,
        *
        enterprise_id: str,
        team_id: str,
        is_enterprise_install: bool
    ):
        return Installation(**self.bot_collection.find_one({
            "enterprise_id": enterprise_id,
            "team_id": team_id,
            "is_enterprise_install": is_enterprise_install
        })["installation"])
        
    def delete_installation(self, *, enterprise_id: str, team_id: str):
        self.installation_collection.delete_one({
            "enterprise_id": enterprise_id,
            "team_id": team_id,
        })

    def delete_bot(self, *, enterprise_id: str, team_id: str, user_id: str):
        self.installation_collection.delete_one({
            "enterprise_id": enterprise_id,
            "team_id": team_id,
            "user_id": user_id
        })
    

class SlackSlashCommandRequest(BaseModel):
    channel_name: str
    channel_id: str
    text: str
    user_name: str
    user_id: str
    team_id: str
    is_enterprise_install: bool
    enterprise_id: str | None = None

@stub.function(
    image=image,
    secrets=secrets,
    timeout=15 * 60,
)
def reply_slack(request: SlackSlashCommandRequest, thread_ts: str | None = None):
    try:
        create_pr = modal.Function.lookup(API_NAME, "create_pr")
        client = None
        installation_store = MongoDBInstallationStore()
        token = installation_store.find_installation(
            team_id=request.team_id,
            enterprise_id=request.enterprise_id,
            user_id=request.user_id,
            is_enterprise_install=request.is_enterprise_install
        ).bot_token
        client = slack_sdk.WebClient(token=token)
    except Exception as e:
        logger.error(f"Error initializing Slack client: {e}")
        raise e
    
    try:
        channel_info = client.conversations_info(channel=request.channel_id)
        channel_description: str = channel_info['channel']['purpose']['value']
        logger.info(f"Channel description: {channel_description}")

        repo_full_name = channel_description.split()[-1]
        logger.info(f"Repo name: {repo_full_name}")

        organization_name, repo_name = repo_full_name.split("/")
        
        try:
            installation_id = get_installation_id(organization_name)
            g = get_github_client(installation_id)
            repo = g.get_repo(repo_full_name)
        except Exception as e:
            # TODO: provide better instructions for installation
            client.chat_postMessage(
                channel=request.channel_id,
                text=f"An error has occurred with fetching the credentials for {repo_full_name}. Please ensure that the app is installed on the Github repo.",
            )
            raise e

        sweep_bot = SweepBot(repo=repo)

        search_already_done = bool(thread_ts)
        if not thread_ts:
            thread = client.chat_postMessage(
                channel=request.channel_id,
                text=f">{request.text}\n- <@{request.user_name}>",
            )
            thread_ts = thread["ts"]
            query = request.text
        else:
            messages = client.conversations_replies(
                channel=request.channel_id,
                ts=thread_ts,
            )["messages"]
            query = thread_query_format.format(messages="\n".join([f"<message user={message['user']}>\n{message['text']}\n</message>" for message in messages]))
    except Exception as e:
        client.chat_postMessage(
            channel=request.channel_id,
            text=":exclamation: Sorry, something went wrong. Sometimes this is because the Github app is not installed on your repo.",
        )
        raise e

    try:
        if not search_already_done:
            logger.info("Fetching relevant snippets...")
            searching_message = client.chat_postMessage(
                channel=request.channel_id,
                text=":mag_right: Searching for relevant snippets...",
                thread_ts=thread_ts
            )
            snippets = sweep_bot.search_snippets(
                request.text,
                installation_id=installation_id
            )
            message = ":mag_right: Some relevant snippets I found:\n\n"
            message += "\n".join(f"{snippet.get_slack_link(repo_name)}\n```{snippet.get_preview()}```" for snippet in snippets)
            client.chat_update(
                channel=request.channel_id,
                ts=searching_message["ts"],
                text=message,
            )
        else:
            snippets = []
        prompt = slack_slash_command_prompt.format(
            relevant_snippets="\n".join([snippet.xml for snippet in snippets]),
            relevant_directories="\n".join([snippet.file_path for snippet in snippets]),
            repo_name=repo_name,
            repo_description=repo.description,
            username=request.user_name,
            query=query
        )
        response = sweep_bot.chat(prompt, functions=functions, function_name={"name": "create_pr"})
        logger.info(response)

        while sweep_bot.messages[-1].function_call is not None:
            obj = sweep_bot.messages[-1].function_call
            name = obj["name"]
            arguments = json.loads(obj["arguments"])
            if name == "get_relevant_snippets":
                logger.info("Searching for relevant snippets...")
                search_message = client.chat_postMessage(
                    channel=request.channel_id,
                    text=f":mag_right: Searching \"{arguments['query']}\" in the codebase...",
                    thread_ts=thread_ts
                )
                additional_snippets = sweep_bot.search_snippets(
                    arguments["query"],
                    installation_id=installation_id
                )
                # additional_snippets = default_snippets
                additional_snippets_message = f":mag_right: Found {len(additional_snippets)} additional snippets with the query \"{arguments['query']}\":\n\n" +  "\n".join(
                    f"{snippet.get_slack_link(repo_name)}\n```{snippet.get_preview()}```" for snippet in additional_snippets
                )
                client.chat_update(
                    channel=request.channel_id,
                    text=additional_snippets_message,
                    ts=search_message["ts"],
                )
                response = sweep_bot.chat(additional_snippets_message, functions=functions)
            elif name == "create_pr":
                title = arguments["title"]
                summary = arguments["summary"]
                branch = arguments["branch"]
                plan = arguments["plan"]
                plan_message = "\n".join(f"• `{file['file_path']}`: {file['instructions']}" for file in plan)
                plan_message = ">" + plan_message.replace("\n", "\n> ")

                creating_pr_message = client.chat_postMessage(
                    channel=request.channel_id,
                    text=pr_format.format(
                        title=title,
                        summary=summary,
                        plan=plan_message
                    ),
                    thread_ts=thread_ts
                )
                file_change_requests = []
                for file in plan:
                    change_type = "create"
                    try:
                        contents = repo.get_contents(file["file_path"])
                        if contents:
                            change_type = "modify"
                    except:
                        pass
                    file_change_requests.append(
                        FileChangeRequest(
                            filename=file["file_path"],
                            instructions=file["instructions"],
                            change_type=change_type
                        )
                    )
                pull_request = PullRequest(
                    title=title,
                    branch_name=branch,
                    content=summary,
                )
                results = create_pr.call(
                    file_change_requests=file_change_requests,
                    pull_request=pull_request,
                    sweep_bot=sweep_bot,
                    username=request.user_name,
                    installation_id=installation_id,
                )
                logger.debug(results)
                pr = results["pull_request"]
                client.chat_update(
                    channel=request.channel_id,
                    text=pr_done_format.format(
                        url=pr.html_url,
                        title=title,
                        summary=summary,
                    ),
                    ts=creating_pr_message["ts"],
                )
                break
            else:
                raise Exception(f"Unknown function {name}")
        else:
            # no breaks were called
            client.chat_postMessage(
                channel=request.channel_id,
                text=response,
                thread_ts=thread_ts
            )
    except Exception as e:
        client.chat_postMessage(
            channel=request.channel_id,
            text=":exclamation: Sorry, something went wrong.",
            thread_ts=thread_ts
        )
        logger.error(f"Error creating PR: {e}")
        raise e


@stub.function(
    image=image,
    keep_warm=1
)
@modal.web_endpoint(method="POST")
async def entrypoint(request: Request):
    body = await request.form()
    request = SlackSlashCommandRequest(**dict(body))
    reply_slack.spawn(request)
    return Response(status_code=200)


@stub.function(image=image, keep_warm=1)
@modal.web_endpoint(method="GET")
def install():
    return RedirectResponse(url=slack_install_link)

def get_oauth_settings():
    from slack_bolt.oauth.oauth_settings import OAuthSettings

    return OAuthSettings(
        client_id=os.environ["SLACK_CLIENT_ID"],
        client_secret=os.environ["SLACK_CLIENT_SECRET"],
        scopes=[
            "app_mentions:read",
            "channels:history",
            "channels:read",
            "chat:write",
            "chat:write.customize",
            "chat:write.public",
            "commands",
            "groups:read",
            "im:read",
            "users.profile:read",
            "users:read",
            "incoming-webhook",
            "mpim:read",
        ],
        install_page_rendering_enabled=False,
        installation_store=MongoDBInstallationStore(),
    )

@stub.function(
    image=image,
    secrets=secrets,
    keep_warm=1
)
@modal.asgi_app(label=PREFIX + "-slack-bot")
def _asgi_app():
    from slack_bolt import App
    from slack_bolt.adapter.fastapi import SlackRequestHandler

    slack_app = App(oauth_settings=get_oauth_settings())

    fastapi_app = FastAPI()
    handler = SlackRequestHandler(slack_app)

    @slack_app.event("url_verification")
    def handle_url_verification(body, logger):
        challenge = body.get("challenge")
        return {"challenge": challenge}

    @slack_app.event("app_mention")
    def handle_app_mentions(body, say, client):
        print("here!")
    
    @slack_app.event("message")
    def handle_message(body, ack, message, client):
        ack()
        print(message)
        if "thread_ts" in message:
            # checking if the message is in a thread
            conversation = client.conversations_replies(
                channel=message["channel"], 
                ts=message["thread_ts"]
            )
            thread_messages = conversation["messages"]
            if thread_messages:
                # checking if the thread has non-zero messages
                bot_profile = thread_messages[0].get("bot_profile")
                if thread_messages[-1].get("bot_profile"):
                    # prevent bots from replying to themselves
                    return
                if bot_profile and bot_profile.get("name") == "Sweep":
                    # checking that the message is from Sweep
                    channel_name = client.conversations_info(channel=message["channel"])["channel"]["name"]
                    reply_slack.call(
                        SlackSlashCommandRequest(
                            channel_id=message["channel"],
                            channel_name=channel_name,
                            text=message["text"],
                            user_name=message["user"],
                            user_id=message["user"],
                            team_id=message["team"],
                            is_enterprise_install=False,
                        ),
                        thread_ts=message["thread_ts"]
                    )
    
    @slack_app.command("/sweep")
    def sweep(ack, respond, command, client):
        print(ack, respond, command, client)
        ack()
        respond("I'm working on it!")
        reply_slack.spawn(SlackSlashCommandRequest(**command))

    @fastapi_app.post("/")
    async def root(request: Request):
        return await handler.handle(request)

    @fastapi_app.get("/slack/install")
    async def oauth_start(request: Request):
        return await handler.handle(request)

    @fastapi_app.get("/slack/oauth_redirect")
    async def oauth_callback(request: Request):
        return await handler.handle(request)

    return fastapi_app



"""
Take a PR and provide an AI generated review of the PR.
"""
from loguru import logger
from sweepai.core.entities import DiffSummarization, PullRequestComment
from sweepai.core.prompts import review_prompt
from sweepai.core.sweep_bot import SweepBot

from sweepai.utils.github_utils import get_file_contents
from sweepai.utils.prompt_constructor import HumanMessageFinalPRComment, HumanMessagePromptReview, HumanMessageReviewFollowup

# Plan:
# 1. Get PR
# 2. Get files changed
# 3. Come up with some comments for the PR
# 4. Take comments and add them to the PR

def get_pr_diffs(repo, pr):
    base_sha = pr.base.sha
    head_sha = pr.head.sha

    comparison = repo.compare(base_sha, head_sha)
    file_diffs = comparison.files

    pr_diffs = []
    for file in file_diffs:
        print(file.status)
        diff = file.patch
        if file.status == "added":
            pr_diffs.append((file.filename, get_file_contents(repo, file_path=file.filename, ref=head_sha), "", diff))
        elif file.status == "modified":
            pr_diffs.append((file.filename, get_file_contents(repo, file_path=file.filename, ref=head_sha), get_file_contents(repo, file_path=file.filename, ref=base_sha), diff))
        elif file.status == "removed":
            pr_diffs.append((file.filename, "", get_file_contents(repo, file_path=file.filename, ref=base_sha), diff))
        else:
            logger.info(f"File status {file.status} not recognized") #TODO(sweep): We don't handle renamed files
    return pr_diffs

def review_pr(repo, pr, issue_url, username, repo_description, title, summary, replies_text, tree):
    repo_name = repo.full_name
    logger.info("Getting PR diffs...")
    diffs = get_pr_diffs(repo, pr)
    human_message = HumanMessagePromptReview(
        repo_name=repo_name,
        issue_url=issue_url,
        username=username,
        repo_description=repo_description,
        title=title,
        summary=summary + replies_text,
        snippets=[],
        tree=tree,
        diffs=[diffs[0] if len(diffs) > 0 else ""],
        pr_title=pr.title,
        pr_message=pr.body or "",
    )
    summarization_replies = []
    sweep_bot = SweepBot.from_system_message_content(
        # human_message=human_message, model="claude-v1.3-100k", repo=repo, is_reply=False
        human_message=human_message, repo=repo, is_reply=False
    )
    summarization_reply = sweep_bot.chat(review_prompt, message_key="review")
    extracted_summary = DiffSummarization.from_string(summarization_reply)
    summarization_replies.append(extracted_summary.content)
    for diff in diffs[1:]:
        review_message = HumanMessageReviewFollowup(diff=diff)
        review_prompt_constructed = review_message.construct_prompt()
        summarization_reply = sweep_bot.chat(review_prompt_constructed, message_key="review")
        extracted_summary = DiffSummarization.from_string(summarization_reply)
        summarization_replies.append(extracted_summary.content)
    final_review_prompt = HumanMessageFinalPRComment(summarization_replies=summarization_replies).construct_prompt()
    reply = sweep_bot.chat(final_review_prompt, message_key="final_review")
    review_comment = PullRequestComment.from_string(reply)
    pr.create_review(body=review_comment.content, event="COMMENT", comments=[])
    changes_required = 'yes' in review_comment.changes_required.lower()
    return changes_required, review_comment.content



"""
Proxy for the UI.
"""

import json
import fastapi
from github import Github
import modal
from loguru import logger
from fastapi import FastAPI
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
from sweepai.app.config import SweepChatConfig

from sweepai.core.chat import ChatGPT
from sweepai.core.entities import FileChangeRequest, Function, Message, PullRequest, Snippet
from sweepai.core.sweep_bot import SweepBot
from sweepai.utils.constants import API_NAME, BOT_TOKEN_NAME, DB_NAME, PREFIX
from sweepai.utils.github_utils import get_github_client, get_installation_id
from sweepai.core.prompts import gradio_system_message_prompt

get_relevant_snippets = modal.Function.from_name(DB_NAME, "get_relevant_snippets")

stub = modal.Stub(PREFIX + "-ui")
image = (
    modal.Image.debian_slim()
    .apt_install("git")
    .pip_install(
        "loguru",
        "tqdm",
        "posthog",
        "openai",
        "anthropic",
        "highlight-io",
        "PyGithub",
        "GitPython",
        "config-path",
        "pyyaml",
    )
)
secrets = [
    modal.Secret.from_name(BOT_TOKEN_NAME),
    modal.Secret.from_name("openai-secret")
]

FUNCTION_SETTINGS = {
    "image": image,
    "secrets": secrets,
    "timeout": 15 * 60,
    "keep_warm": 1
}

functions = [
    Function(
        name="create_pr",
        description="Creates a PR.",
        parameters={
            "properties": {
                "plan": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "file_path": {
                                "type": "string",
                                "description": "The file path to change."
                            },
                            "instructions": {
                                "type": "string",
                                "description": "Concise NATURAL LANGUAGE summary of what to change in each file. There should be absolutely NO code.",
                                "example":  [
                                    "Refactor the algorithm by moving the main function to the top of the file.",
                                    "Change the implementation to recursion"
                                ]
                            },
                        },
                        "required": ["file_path", "instructions"]
                    },
                    "description": "A list of files to modify or create and corresponding instructions."
                },
                "title": {
                    "type": "string",
                    "description": "Title of PR",
                },
                "summary": {
                    "type": "string",
                    "description": "Detailed summary of PR",
                },
                "branch": {
                    "type": "string",
                    "description": "Name of branch to create PR in.",
                },
            },
            "required": ["plan", "title", "summary", "branch"]
        }
    ),
]

@stub.function(**FUNCTION_SETTINGS)
@modal.asgi_app(label=PREFIX + "-ui")
def _asgi_app():
    app = FastAPI()

    def verify_config(request: SweepChatConfig) -> bool:
        try:
            github_user_client = Github(request.github_pat)
            repo = github_user_client.get_repo(request.repo_full_name)
            assert repo
        except Exception as e:
            logger.warning(e)
            raise fastapi.HTTPException(status_code=403, detail="You do not have access to this repo")
        return True

    @app.post("/installation_id")
    def installation_id(request: SweepChatConfig) -> dict:
        # first check if user has access to the repo
        assert verify_config(request)

        try:
            organization, _repo_name = request.repo_full_name.split("/")
            installation_id = get_installation_id(organization)
            assert installation_id
        except Exception as e:
            logger.warning(e)
            raise fastapi.HTTPException(status_code=403, detail="Sweep app is not installed on this repo. To install it, go to https://github.com/apps/sweep-ai")

        return {"installation_id": installation_id}

    class SearchRequest(BaseModel):
        query: str
        config: SweepChatConfig
        n_results: int = 5

    @app.post("/search")
    def search(request: SearchRequest) -> list[Snippet]:
        logger.info("Searching for snippets...")
        get_relevant_snippets = modal.Function.lookup(DB_NAME, "get_relevant_snippets")

        assert verify_config(request.config)

        snippets: list[Snippet] = get_relevant_snippets.call(
            request.config.repo_full_name,
            request.query,
            n_results=request.n_results,
            installation_id=request.config.installation_id
        )
        g = get_github_client(request.config.installation_id)
        repo = g.get_repo(request.config.repo_full_name)
        for snippet in snippets:
            try:
                snippet.content = repo.get_contents(snippet.file_path).decoded_content.decode("utf-8")
            except Exception:
                logger.error(snippet)
        return snippets

    class CreatePRRequest(BaseModel):
        # proposed PR information
        file_change_requests: list[tuple[str, str]]
        pull_request: PullRequest 

        # state information
        messages: list[tuple[str | None, str | None]]
        snippets: list[Snippet] = []

        config: SweepChatConfig
    
    @app.post("/create_pr")
    def create_pr(request: CreatePRRequest):
        assert verify_config(request.config)

        g = get_github_client(request.config.installation_id)
        repo = g.get_repo(request.config.repo_full_name)

        create_pr_func = modal.Function.lookup(API_NAME, "create_pr")
        system_message = gradio_system_message_prompt.format(
            snippets="\n".join([snippet.denotation + f"\n```{snippet.get_snippet()}```" for snippet in request.snippets]),
            repo_name=request.config.repo_full_name,
            repo_description="Sweep is an AI junior developer"
        )

        def file_exists(file_path: str) -> bool:
            try:
                repo.get_contents(file_path)
                return True
            except Exception:
                return False

        results = create_pr_func.call(
            [FileChangeRequest(
                filename = item[0],
                instructions = item[1],
                change_type = "modify" if file_exists(item[0]) else "create", # TODO update this
            ) for item in request.file_change_requests],
            request.pull_request,
            SweepBot(
                repo = repo,
                messages = [Message(role="system", content=system_message, key="system")] +
                    [Message.from_tuple(message) for message in request.messages],
            ),
            request.config.github_username,
            installation_id = request.config.installation_id,
            issue_number = None,
        )
        generated_pull_request = results["pull_request"]
        print(generated_pull_request)
        return {
            "html_url": generated_pull_request.html_url,
        }
    
    class ChatRequest(BaseModel):
        messages: list[tuple[str | None, str | None]]
        snippets: list[Snippet]
        config: SweepChatConfig

    @app.post("/chat")
    def chat(
        request: ChatRequest,
    ) -> str:
        assert verify_config(request.config)

        messages = [Message.from_tuple(message) for message in request.messages]
        chatgpt = ChatGPT(messages=messages[:-1])
        result = chatgpt.chat(messages[-1].content, model="gpt-4-0613")
        return result
    
    @app.post("/chat_stream")
    def chat_stream(request: ChatRequest):
        assert verify_config(request.config)

        messages = [Message.from_tuple(message) for message in request.messages]
        system_message = gradio_system_message_prompt.format(
            snippets="\n".join([snippet.denotation + f"\n```{snippet.get_snippet()}```" for snippet in request.snippets]),
            repo_name=request.config.repo_full_name,
            repo_description="" # TODO: fill this
        )
        chatgpt = ChatGPT(messages=[Message(role="system", content=system_message, key="system")] + messages[:-1])
        return StreamingResponse(
            (json.dumps(chunk) for chunk in chatgpt.chat_stream(messages[-1].content, model="gpt-4-0613", functions=functions)),
            media_type="text/event-stream"
        )
    return app



from typing import Literal
from pydantic import BaseModel


class Installation(BaseModel):
    id: str

class InstallationCreatedRequest(BaseModel):
    class Repository(BaseModel):
        full_name: str
    repositories: list[Repository]
    installation: Installation

class ReposAddedRequest(BaseModel):
    class Repository(BaseModel):
        full_name: str
    repositories_added: list[Repository]
    installation: Installation

class CommentCreatedRequest(BaseModel):
    class Comment(BaseModel):
        class User(BaseModel):
            login: str
        body: str | None
        original_line: int
        path: str
        diff_hunk: str
        user: User

    class PullRequest(BaseModel):
        class Head(BaseModel):
            ref: str
        number: int
        body: str | None
        state: str  # "closed" or "open"
        head: Head
        title: str

    class Repository(BaseModel):
        full_name: str
        description: str | None

    class Sender(BaseModel):
        pass

    action: str
    comment: Comment
    pull_request: PullRequest
    repository: Repository
    sender: Sender
    installation: Installation


class IssueRequest(BaseModel):
    class Issue(BaseModel):
        class User(BaseModel):
            login: str
            type: str

        class Assignee(BaseModel):
            login: str

        class Repository(BaseModel):
            # TODO(sweep): Move this out
            full_name: str
            description: str | None
        
        class Label(BaseModel):
            name: str

        class PullRequest(BaseModel):
            pass

        pull_request: PullRequest | None
        title: str
        number: int
        html_url: str
        user: User
        body: str | None
        labels: list[Label]
        assignees: list[Assignee]

    action: str
    issue: Issue
    repository: Issue.Repository
    assignee: Issue.Assignee | None
    installation: Installation

class IssueCommentRequest(IssueRequest):
    class Comment(BaseModel):
        class User(BaseModel):
            login: str
            type: Literal["User", "Bot"]
        user: User
        id: int
        body: str
    comment: Comment

class PRRequest(BaseModel):
    class PullRequest(BaseModel):
        class User(BaseModel):
            login: str
        class MergedBy(BaseModel):
            login: str
        user: User
        merged_by: MergedBy
    class Repository(BaseModel):
        full_name: str
    pull_request: PullRequest
    repository: Repository


import json
from github import Github
import gradio as gr
from loguru import logger
import webbrowser

from sweepai.app.api_client import APIClient
from sweepai.app.config import SweepChatConfig
from sweepai.core.entities import Snippet

config = SweepChatConfig.load()

api_client = APIClient(config=config)

pr_summary_template = """💡 I'll create the following PR:

**{title}**
{summary}

Here is my plan:
{plan}

Reply with "ok" to create the PR or anything else to propose changes."""

github_client = Github(config.github_pat)
repos = github_client.get_user().get_repos()

with gr.Blocks(theme=gr.themes.Soft(), title="Sweep Chat", css="footer {visibility: hidden;}") as demo:
    repo_full_name = gr.Dropdown(choices=[repo.full_name for repo in repos], label="Repo full name", value=config.repo_full_name or "")
    with gr.Row():
        with gr.Column(scale=2):
            chatbot = gr.Chatbot(height=650)
        with gr.Column():
            snippets_text = gr.Markdown(value="### Relevant snippets")
    msg = gr.Textbox(label="Message to Sweep", placeholder="Write unit tests for OpenAI calls")
    clear = gr.ClearButton([msg, chatbot, snippets_text])

    snippets: list[Snippet] = []
    proposed_pr: str | None = None

    def repo_name_change(repo_full_name):
        # check if user has access to the repo and if the app is installed on the repo
        global installation_id
        try:
            config.repo_full_name = repo_full_name
            api_client.config = config
            installation_id = api_client.get_installation_id()
            assert installation_id
            config.installation_id = installation_id
            api_client.config = config
            config.save()
            return ""
        except Exception as e:
            webbrowser.open_new_tab("https://github.com/apps/sweep-ai")
            config.repo_full_name = None
            config.installation_id = None
            config.save()
            api_client.config = config
            raise e

    repo_full_name.change(repo_name_change, [repo_full_name], [msg])

    def handle_message_submit(repo_full_name: str, user_message: str, history: list[tuple[str | None, str | None]]):
        if not repo_full_name:
            raise Exception("Set the repository name first")
        return gr.update(value="", interactive=False), history + [[user_message, None]]

    def handle_message_stream(chat_history: list[tuple[str | None, str | None]], snippets_text, repo_name):
        snippets = []
        yield chat_history, snippets_text
        if snippets_text and snippets_text.strip().count("\n") == 0:
            # Searching for relevant snippets
            chat_history[-1][1] = "Searching for relevant snippets..."
            yield chat_history, snippets_text
            logger.info("Fetching relevant snippets...")
            snippets = api_client.search(chat_history[-1][0], 5)
            logger.info("Fetched relevant snippets.")
            chat_history[-1][1] = "Found relevant snippets."
            
            backtick, escaped_backtick = "`", "\\`"
            snippets_text = "### Relevant snippets:\n" + "\n".join([f"{snippet.denotation}\n```python\n{snippet.get_preview(5).replace(backtick, escaped_backtick)}\n```" for snippet in snippets])
            yield chat_history, snippets_text
        
        global proposed_pr
        if proposed_pr and chat_history[-1][0].strip().lower() in ("okay", "ok"):
            chat_history[-1][1] = f"⏳ Creating PR..."
            yield chat_history, snippets_text
            pull_request = api_client.create_pr(
                file_change_requests=[(item["file_path"], item["instructions"]) for item in proposed_pr["plan"]],
                pull_request={
                    "title": proposed_pr["title"],
                    "content": proposed_pr["summary"],
                    "branch_name": proposed_pr["branch"],
                },
                messages=chat_history,
            )
            chat_history[-1][1] = f"✅ PR created at {pull_request['html_url']}"
            yield chat_history, snippets_text
            return

        # Generate response
        logger.info("...")
        chat_history.append([None, "Fetching endpoint..."])
        yield chat_history, snippets_text
        chat_history[-1][1] = ""
        logger.info("Starting to generate response...")
        stream = api_client.stream_chat(chat_history, snippets)
        function_name = ""
        raw_arguments = ""
        for chunk in stream:
            if chunk.get("content"):
                token = chunk["content"]
                chat_history[-1][1] += token
                yield chat_history, snippets_text
            if chunk.get("function_call"):
                function_call = chunk["function_call"]
                function_name = function_name or function_call.get("name")
                raw_arguments += function_call.get("arguments")
                chat_history[-1][1] = f"Calling function: `{function_name}`\n```json\n{raw_arguments}\n```"
                yield chat_history, snippets_text
        if function_name:
            arguments = json.loads(raw_arguments)
            if function_name == "create_pr":
                assert "title" in arguments
                assert "summary" in arguments
                assert "plan" in arguments
                if "branch" not in arguments:
                    arguments["branch"] = arguments["title"].lower().replace(" ", "_").replace("-", "_")[:50]
                chat_history[-1][1] = pr_summary_template.format(
                    title=arguments["title"],
                    summary=arguments["summary"],
                    plan="\n".join([f"* `{item['file_path']}`: {item['instructions']}" for item in arguments["plan"]])
                )
                yield chat_history, snippets_text
                proposed_pr = arguments
            else:
                raise NotImplementedError

    response = msg.submit(handle_message_submit, [repo_full_name, msg, chatbot], [msg, chatbot], queue=False).then(handle_message_stream, [chatbot, snippets_text, repo_full_name], [chatbot, snippets_text])
    response.then(lambda: gr.update(interactive=True), None, [msg], queue=False)


if __name__ == "__main__":
    demo.queue()
    demo.launch()



import typer

from sweepai.app.config import SweepChatConfig

epilog = "Sweep is a AI junior developer. Docs at https://docs.sweep.dev, install at https://github.com/apps/sweep-ai and support at https://discord.gg/sweep-ai."

app = typer.Typer(epilog=epilog)

@app.callback(invoke_without_command=True)
def start():
    """
    Launch Sweep Chat in the browser
    """
    SweepChatConfig.load()
    from sweepai.app.ui import demo
    demo.queue()
    demo.launch(enable_queue=True, inbrowser=True)
    
@app.command()
def auth():
    """
    Reauthenticate with Github API for Sweep to work (for token expiry)
    """
    SweepChatConfig.load(recreate=True)
    print("Setup completed successfully!")


if __name__ == "__main__":
    app()



import json
from loguru import logger
import github
from github.Repository import Repository
from github.ContentFile import ContentFile
from github.GithubException import GithubException
import modal
from pydantic import BaseModel
from sweepai.core.code_repair import CodeRepairer

from sweepai.core.entities import (
    FileChange,
    FileChangeRequest,
    FilesToChange,
    PullRequest,
    RegexMatchError,
    Function,
    Snippet
)
from sweepai.core.chat import ChatGPT
from sweepai.core.prompts import (
    files_to_change_prompt,
    pull_request_prompt,
    create_file_prompt,
    modify_file_prompt,
    modify_file_plan_prompt,
)
from sweepai.utils.constants import DB_NAME
from sweepai.utils.diff import format_contents, generate_diff, generate_new_file, is_markdown


class CodeGenBot(ChatGPT):

    def get_files_to_change(self):
        file_change_requests: list[FileChangeRequest] = []
        for count in range(5):
            try:
                logger.info(f"Generating for the {count}th time...")
                files_to_change_response = self.chat(files_to_change_prompt, message_key="files_to_change") # Dedup files to change here
                files_to_change = FilesToChange.from_string(files_to_change_response)
                files_to_create: list[str] = files_to_change.files_to_create.split("*")
                files_to_modify: list[str] = files_to_change.files_to_modify.split("*")
                for file_change_request, change_type in zip(
                    files_to_create + files_to_modify,
                    ["create"] * len(files_to_create)
                    + ["modify"] * len(files_to_modify),
                ):
                    file_change_request = file_change_request.strip()
                    if not file_change_request or file_change_request == "None":
                        continue
                    logger.debug(file_change_request)
                    logger.debug(change_type)
                    file_change_requests.append(
                        FileChangeRequest.from_string(
                            file_change_request, change_type=change_type
                        )
                    )
                if file_change_requests:
                    return file_change_requests
            except RegexMatchError:
                logger.warning("Failed to parse! Retrying...")
                self.delete_messages_from_chat("files_to_change")
                continue
        raise Exception("Could not generate files to change")

    def generate_pull_request(self):
        pull_request = None
        for count in range(5):
            try:
                logger.info(f"Generating for the {count}th time...")
                pr_text_response = self.chat(pull_request_prompt, message_key="pull_request")
            except Exception:
                logger.warning("Failed to parse! Retrying...")
                self.undo()
                continue
            pull_request = PullRequest.from_string(pr_text_response)
            pull_request.branch_name = "sweep/" + pull_request.branch_name[:250]
            return pull_request
        raise Exception("Could not generate PR text")


class GithubBot(BaseModel):
    class Config:
        arbitrary_types_allowed = True  # for repo: Repository

    repo: Repository

    def get_contents(self, path: str, branch: str = ""):
        if not branch:
            branch = self.repo.default_branch
        try:
            return self.repo.get_contents(path, ref=branch)
        except Exception as e:
            logger.error(path)
            raise e

    def get_file(self, file_path: str, branch: str = "") -> ContentFile:
        content = self.get_contents(file_path, branch)
        assert not isinstance(content, list)
        return content

    def check_path_exists(self, path: str, branch: str = ""):
        try:
            self.get_contents(path, branch)
            return True
        except Exception:
            return False

    def create_branch(self, branch: str) -> str:
        # Generate PR if nothing is supplied maybe
        base_branch = self.repo.get_branch(self.repo.default_branch)
        try:
            self.repo.create_git_ref(f"refs/heads/{branch}", base_branch.commit.sha)
            return branch
        except GithubException as e:
            logger.error(f"Error: {e}, trying with other branch names...")
            for i in range(1, 100):
                try:
                    logger.warning(f"Retrying {branch}_{i}...")
                    self.repo.create_git_ref(
                        f"refs/heads/{branch}_{i}", base_branch.commit.sha
                    )
                    return f"{branch}_{i}"
                except GithubException:
                    pass
            raise e
    
    def populate_snippets(self, snippets: list[Snippet]):
        for snippet in snippets:
            try:
                snippet.content = self.repo.get_contents(snippet.file_path).decoded_content.decode("utf-8")
            except Exception as e:
                logger.error(snippet)
    
    def search_snippets(
        self, 
        query: str, 
        installation_id: str,
        num_snippets: int = 30,
    ) -> list[Snippet]:
        get_relevant_snippets = modal.Function.lookup(DB_NAME, "get_relevant_snippets")
        snippets: list[Snippet] = get_relevant_snippets.call(
            self.repo.full_name, 
            query=query,
            n_results=num_snippets,
            installation_id=installation_id,
        )
        self.populate_snippets(snippets)
        return snippets


class SweepBot(CodeGenBot, GithubBot):
    def cot_retrieval(self):
        # TODO(sweep): add semantic search using vector db
        # TODO(sweep): add search using webpilot + github
        functions = [
            Function(
                name="cat",
                description="Cat files. Max 3 files per request.",
                parameters={
                    "properties": {
                        "filepath": {
                            "type": "string",
                            "description": "Paths to files. One per line."
                        },
                    }
                } # manage file too large
            ),
            Function(
                name="finish",
                description="Indicate you have sufficient data to proceed.",
                parameters={"properties": {}} 
            ),
        ]


        # self.chat(
        #     cot_retrieval_prompt,
        #     message_key="cot_retrieval",
        #     functions=functions,
        # )
        # is_function_call = self.messages[-1].function_call is not None
        # for _retry in range(3):
        #     logger.info("Got response.")
        #     if not is_function_call:
        #         break

        #     response = self.messages[-1].function_call
        #     # response = json.loads(response)
        #     function_name = response["name"]
        #     arguments = response["arguments"]
        #     logger.info(f"Fetching file {function_name} with arguments {arguments}.")
        #     arguments = json.loads(arguments)
        #     if function_name == "finish":
        #         return
        #     elif function_name == "cat":
        #         path = arguments["filepath"]
        #         try:
        #             logger.info("Retrieving file...")
        #             content = self.get_file(path).decoded_content.decode("utf-8")
        #             logger.info("Received file")
        #         except github.GithubException:
        #             response = self.chat(
        #                 f"File not found: {path}",
        #                 message_key=path,
        #                 functions=functions,
        #             )
        #         else:
        #             response = self.chat(
        #                 f"Here is the file: <file path=\"{path}\">\n\n{content[:10000]}</file>. Fetch more content or call finish.", 
        #                 message_key=path,
        #                 functions=functions
        #             ) # update this constant
        return

    def create_file(self, file_change_request: FileChangeRequest) -> FileChange:
        file_change: FileChange | None = None
        for count in range(5):
            create_file_response = self.chat(
                create_file_prompt.format(
                    filename=file_change_request.filename,
                    instructions=file_change_request.instructions,
                ),
                message_key=f"file_change_{file_change_request.filename}",
            )
            # Add file to list of changed_files
            self.file_change_paths.append(file_change_request.filename)
            # self.delete_file_from_system_message(file_path=file_change_request.filename)
            try:
                file_change = FileChange.from_string(create_file_response)
                assert file_change is not None
                file_change.commit_message = f"sweep: {file_change.commit_message[:50]}"
                return file_change
            except Exception:
                logger.warning(f"Failed to parse. Retrying for the {count}th time...")
                self.undo()
                continue
        raise Exception("Failed to parse response after 5 attempts.")

    def modify_file(
        self, file_change_request: FileChangeRequest, contents: str = ""
    ) -> tuple[str, str]:
        if not contents:
            contents = self.get_file(
                file_change_request.filename
            ).decoded_content.decode("utf-8")
        # Add line numbers to the contents; goes in prompts but not github
        contents_line_numbers = "\n".join([f"{i}:{line}" for i, line in enumerate(contents.split("\n"))])
        contents_line_numbers = contents_line_numbers.replace('"""', "'''")
        for count in range(5):
            if "0613" in self.model:
                _ = self.chat( # We don't use the plan in the next call
                    modify_file_plan_prompt.format(
                        filename=file_change_request.filename,
                        instructions=file_change_request.instructions,
                        code=contents_line_numbers,
                    ),
                    message_key=f"file_change_{file_change_request.filename}",
                )
                modify_file_response = self.chat(
                    modify_file_prompt,
                    message_key=f"file_change_{file_change_request.filename}",
                )
                try:
                    logger.info(f"modify_file_response: {modify_file_response}")
                    new_file = generate_new_file(modify_file_response, contents)
                    if not is_markdown(file_change_request.filename):
                        code_repairer = CodeRepairer()
                        diff = generate_diff(old_code=contents, new_code=new_file)
                        new_file = code_repairer.repair_code(diff=diff, user_code=new_file, feature=file_change_request.instructions)
                    return (new_file, file_change_request.filename)
                except Exception as e:
                    logger.warning(f"Recieved error {e}")
                    logger.warning(
                        f"Failed to parse. Retrying for the {count}th time..."
                    )
                    self.undo()
                    self.undo()
                    continue
        raise Exception("Failed to parse response after 5 attempts.")
 
    def change_file(self, file_change_request: FileChangeRequest):
        if file_change_request.change_type == "create":
            return self.create_file(file_change_request)
        elif file_change_request.change_type == "modify":
            return self.create_file(file_change_request)
        else:
            raise Exception("Not a valid file type")

    def change_files_in_github(
        self,
        file_change_requests: list[FileChangeRequest],
        branch: str,
    ):
        # should check if branch exists, if not, create it
        logger.debug(file_change_requests)
        for file_change_request in file_change_requests:
            file_markdown = is_markdown(file_change_request.filename)
            if file_change_request.change_type == "create":
                try: # Try to create
                    file_change = self.create_file(file_change_request)
                    logger.debug(
                        f"{file_change_request.filename}, {file_change.commit_message}, {file_change.code}, {branch}"
                    )
                    file_change.code = format_contents(file_change.code, file_markdown)
                    self.repo.create_file(
                        file_change_request.filename,
                        file_change.commit_message,
                        file_change.code,
                        branch=branch,
                    )
                except github.GithubException as e:
                    logger.info(e)
                    try: # Try to modify
                        contents = self.get_file(file_change_request.filename, branch=branch)
                        file_change.code = format_contents(file_change.code, file_markdown)
                        self.repo.update_file(
                            file_change_request.filename,
                            file_change.commit_message,
                            file_change.code,
                            contents.sha,
                            branch=branch,
                        )
                    except:
                        pass
            elif file_change_request.change_type == "modify":
                # TODO(sweep): Cleanup this
                try:
                    contents = self.get_file(file_change_request.filename, branch=branch)
                except github.UnknownObjectException as e:
                    logger.warning(f"Received error {e}, trying creating file...")
                    file_change_request.change_type = "create"
                    self.create_file(file_change_request)
                    file_change = self.create_file(file_change_request)
                    logger.debug(
                        f"{file_change_request.filename}, {file_change.commit_message}, {file_change.code}, {branch}"
                    )
                    file_change.code = format_contents(file_change.code, file_markdown)
                    self.repo.create_file(
                        file_change_request.filename,
                        file_change.commit_message,
                        file_change.code,
                        branch=branch,
                    )
                else:
                    new_file_contents, file_name = self.modify_file(
                        file_change_request, contents.decoded_content.decode("utf-8")
                    )
                    new_file_contents = format_contents(new_file_contents, file_markdown)
                    logger.debug(
                        f"{file_name}, {f'Update {file_name}'}, {new_file_contents}, {branch}"
                    )
                    self.repo.update_file(
                        file_name,
                        f'Update {file_name}',
                        new_file_contents,
                        contents.sha,
                        branch=branch,
                    )
            else:
                raise Exception("Invalid change type")
